/*
 * Polish Lexical Analyser
 * Stanislaw Galus
 * 1 November 2008
 */

/* $Id: pla.c,v 1.21 2008/11/01 06:14:01 sgalus Exp $ */

/*
 * Compilation:
 *
 * gcc -ansi -pedantic -Wall -W -g -O2 -DMCHECK -static pla.c sl.c
 *   -o pla
 *
 * bcc32 pla.c sl.c -opla
 */

#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef MCHECK                   /* enables memory debugging by GCC */
#include <mcheck.h>
#endif
#include "sl.h"
#include "pla.h"

#define NELEMS(a) ((int)(sizeof(a) / sizeof(a[0])))
#define STMT(stuff) do {stuff} while (0)
#define NEW(type) ((type *)malloc(sizeof(type)))
#define NEW_VECTOR(type, dim) \
  ((type *)malloc((size_t)(dim) * sizeof(type)))

const char *const version = "1.1";

/**************************************************************
 *                                                            *
 *                  Miscellaneous utilities.                  *
 *                                                            *
 **************************************************************/

/* l2_ functions work for the ISO_8859-2:1987 character set. */

#define SPACE 0x01u
#define DIGIT 0x02u
#define UPPER 0x04u
#define LOWER 0x08u
#define HEX 0x10u
#define CNTRL 0x20u
#define PUNCT 0x40u
#define TYPE(c, t)                                            \
c == EOF ? 0                                                  \
         : (int)((unsigned char)type[(int)(unsigned char)(c)] \
                 & (unsigned char)(t))

static const char type[256] =
{
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x21, 0x21, 0x21, 0x21, 0x21, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x01, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
  0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
  0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12, 0x12,
  0x12, 0x12, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
  0x40, 0x14, 0x14, 0x14, 0x14, 0x14, 0x14, 0x04,
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
  0x04, 0x04, 0x04, 0x40, 0x40, 0x40, 0x40, 0x40,
  0x40, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x08,
  0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
  0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
  0x08, 0x08, 0x08, 0x40, 0x40, 0x40, 0x40, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
  0x40, 0x04, 0x40, 0x04, 0x40, 0x04, 0x04, 0x40,
  0x40, 0x04, 0x04, 0x04, 0x04, 0x40, 0x04, 0x04,
  0x40, 0x08, 0x40, 0x08, 0x40, 0x08, 0x08, 0x40,
  0x40, 0x08, 0x08, 0x08, 0x08, 0x40, 0x08, 0x08,
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04,
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x40,
  0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x04, 0x08,
  0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
  0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
  0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x40,
  0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x40
};

int l2_isalnum(int c) {return TYPE(c, LOWER | UPPER | DIGIT);}
int l2_isalpha(int c) {return TYPE(c, LOWER | UPPER);}
int l2_iscntrl(int c) {return TYPE(c, CNTRL);}
int l2_isdigit(int c) {return TYPE(c, DIGIT);}
int l2_isgraph(int c) {return TYPE(c, LOWER | UPPER | DIGIT | PUNCT);}
int l2_islower(int c) {return TYPE(c, LOWER);}
int l2_isprint(int c) {return c == EOF ? 0 : l2_isgraph(c) ||
                         (unsigned char)c == 0x20u;}
int l2_ispunct(int c) {return TYPE(c, PUNCT);}
int l2_isspace(int c) {return TYPE(c, SPACE);}
int l2_isupper(int c) {return TYPE(c, UPPER);}
int l2_isxdigit(int c) {return TYPE(c, HEX);}

int l2_tolower(int c)
{
  if (l2_isupper(c))
    return (unsigned char)c > 0xa0u && (unsigned char)c < 0xc0u
      ? (unsigned char)c + 0x10
      : (unsigned char)c + 0x20;
  return c;
}

int l2_toupper(int c)
{
  if (l2_islower(c) && (unsigned char)c != 0xdfu)
    return (unsigned char)c > 0xa0u && (unsigned char)c < 0xc0u
      ? (unsigned char)c - 0x10
      : (unsigned char)c - 0x20;
  return c;
}

#undef SPACE
#undef DIGIT
#undef UPPER
#undef LOWER
#undef HEX
#undef CNTRL
#undef PUNCT
#undef TYPE

int l2_islower_str(const char *s)
{
  if (!*s)
    return 0;
  while (l2_islower(*s))
    s++;
  return !*s;
}

int l2_isupper_str(const char *s)
{
  if (!*s)
    return 0;
  while (l2_isupper(*s))
    s++;
  return !*s;
}

/*
 * True if and only if the string is not empty, its first character is
 * an upper case letter and the following characters, if they exist, are
 * lower case letters.
 */
int l2_iscapit_str(const char *s)
{
  if (!l2_isupper(*s++))
    return 0;
  while (l2_islower(*s))
    s++;
  return !*s;
}

char *l2_tolower_str(char *s)
{
  char *t;

  for (t = s; *t; t++)
    *t = l2_tolower(*t);
  return s;
}

char *l2_toupper_str(char *s)
{
  char *t;

  for (t = s; *t; t++)
    *t = l2_toupper(*t);
  return s;
}

char *l2_tocapit_str(char *s)
{
  char *t;

  if ((*s = l2_toupper(*s)))
    for (t = s + 1; *t; t++)
      *t = l2_tolower(*t);
  return s;
}

int l2_isvowel(char c)          /* temporary implementation */
{
  return
    strchr("AEIOUYaeiouy\241\261\301\302\303\304\311\312\313\314"
           "\315\316\323\324\325\326\331\332\333\334\335\341\342"
           "\343\344\351\352\353\354\355\356\363\364\365\366\371"
           "\372\373\374\375", c) != NULL;
}

int l2_isconsonant(char c)      /* temporary implementation */
{
  return l2_isalpha(c) && !l2_isvowel(c);
}

char *strdup(const char *s)
{
  size_t n = strlen(s) + 1;
  char *t = malloc(n);
  return t == NULL ? NULL : (char *)memcpy(t, s, n);
}

/*
 * Remove white space from both sides of s and replace every sequence of
 * white space by a single space.
 */
char *trim(char *s)
{
  char *t = s, *u = s;

  while (isspace(*u))
    u++;
  while ((*t++ = *u++))
    if (isspace(*u))
    {
      while (isspace(*++u)) ;
      if (*u)
        *t++ = ' ';
      else
      {
        *t = '\0';
        break;
      }
    }
  return s;
}

/* True if s is a concatenation of t and u. */
int iscat(const char *s, const char *t, const char *u)
{
  for (; *s == *t; s++, t++)
    if (*s == '\0')
      return *u == '\0';
  if (*t != '\0')
    return 0;
  for (; *s == *u; s++, u++)
    if (*s == '\0')
      return 1;
  return 0;
}

/* True if s2 is a prefix of s1. */
int isprefix(const char *s1, const char *s2)
{
  return strncmp(s1, s2, strlen(s2)) == 0;
}

/* True if s2 is a suffix of s1. */
int issuffix(const char *s1, const char *s2)
{
  const size_t l1 = strlen(s1);
  const size_t l2 = strlen(s2);
  return l1 >= l2 && !strcmp(s1 + l1 - l2, s2);
}

/* True if s2 is a suffix of s1 and s1 is longer than s2. */
int ispsuffix(const char *s1, const char *s2)
{
  const size_t l1 = strlen(s1);
  const size_t l2 = strlen(s2);
  return l1 > l2 && !strcmp(s1 + l1 - l2, s2);
}

int ispsuffix2(const char *s1, size_t l1, const char *s2, size_t l2)
{
  return l1 > l2 && !strcmp(s1 + l1 - l2, s2);
}

/*
 * Calculates BSD 16-bit checksum. Source: GNU core utilities, sum.c,
 * ftp://ftp.task.gda.pl/pub/gnu/coreutils/coreutils-6.9.tar.bz2, 9 VIII
 * 2008.
 */
int chksum(const void *ptr, size_t size)
{
  const unsigned char *c = ptr;
  int s = 0;
  while (size-- > 0)
    s = ((s >> 1) + ((s & 1) << 15) + *c++) & 0xffff;
  return s;
}

/* Calculates BSD 16-bit checksum of a file. */
int chksumf(const char *fn)
{
  FILE *f;
  int c, s = 0;

  if ((f = fopen(fn, "r")) == NULL)
    return -1;
  while ((c = getc(f)) != EOF)
    s = ((s >> 1) + ((s & 1) << 15) + c) & 0xffff;
  if (ferror(f))
  {
    fclose(f);
    return -1;
  }
  if (fclose(f) != 0)
    return -2;
  return s;
}

/**************************************************************
 *                                                            *
 *                      Lexical symbol.                       *
 *                                                            *
 **************************************************************/

static const int top1tab[] =    /* uninflected pronouns (T0) */
{
  INDEFINITE, NEGATIVE, DEMONSTRATIVE, INTERROGATIVE_RELATIVE
};

static const int top2tab[] =    /* pronouns inflected substantivally
                                 (TPs) */
{
  PERSONAL
};

static const int top3tab[] =    /* pronouns inflected adjectivally
                                 (TPa) */
{
  PERSONAL, INDEFINITE, NEGATIVE, DEMONSTRATIVE, POSSESSIVE,
  INTERROGATIVE_RELATIVE
};

static const int top4tab[] =    /* pronouns inflected like numerals
                                 (TPn) */
{
  INDEFINITE, DEMONSTRATIVE, INTERROGATIVE_RELATIVE
};

static const int top5tab[] =    /* pronouns inflected singularly by case
                                 (TPc) */
{
  REFLEXIVE, INDEFINITE, NEGATIVE, INTERROGATIVE_RELATIVE
};

static const int tontab[] =     /* types of numerals */
{
  CARDINAL, COLLECTIVE, FRACTIONAL, ORDINAL
};

static const int fovtab[] =     /* forms of verbs */
{
  INFINITIVE, SIMPLE_PRESENT_OR_FUTURE, PAST, IMPERATIVE_MOOD,
  SUBJUNCTIVE_MOOD, IMPERSONAL_FORM_OF_PAST_TENSE, ADVERBIAL_PARTICIPLE
};

static const int i1tab[] =      /* inflexions of pronouns */
{
  0, SUBSTANTIVAL_DECLENSION, ADJECTIVAL_DECLENSION,
  NUMERAL_DECLENSION, SINGULAR_DECLENSION_BY_CASE
};

static const int i2tab[] =      /* inflexions of cardinal numerals */
{
  SUBSTANTIVAL_DECLENSION, NUMERAL_DECLENSION,
  DECLENSION_OF_NUMERAL_1, DECLENSION_OF_NUMERAL_2
};

static const int i3tab[] =      /* inflexions of fractional numerals */
{
  0, SINGULAR_DECLENSION_BY_GENDER
};

static const int ctab[] =       /* cases */
{
  NOMINATIVE, GENITIVE, DATIVE, ACCUSATIVE, INSTRUMENTAL, LOCATIVE,
  VOCATIVE
};

static const int ntab[] =       /* numbers */
{
  SINGULAR, PLURAL
};

static const int g1tab[] =      /* genders of nouns */
{
  0, MASCULINE_PERSONAL, MASCULINE_ANIMAL, MASCULINE_INANIMATE,
  FEMININE, NEUTER
};

static const int g2tab[] =      /* genders of adjectives, singular */
{
  MASCULINE_ANIMATE, MASCULINE_INANIMATE, FEMININE,  NEUTER
};

static const int g3tab[] =      /* genders of adjectives, plural, also
                                 used in conjugation, plural */
{
  MASCULINE_PERSONAL, NON_MASCULINE_PERSONAL
};

static const int g4tab[] =      /* genders of numeral declension */
{
  MASCULINE_PERSONAL, NON_MASCULINE_PERSONAL
};

static const int g5tab[] =      /* genders of declension of numeral 1 */
{
  MASCULINE_ANIMATE, MASCULINE_INANIMATE, FEMININE, NEUTER
};

static const int g6tab[] =      /* genders of declension of numeral 2 */
{
  MASCULINE_IMPERSONAL, MASCULINE_PERSONAL, FEMININE, NEUTER
};

static const int g7tab[] =      /* genders of singular declension by
                                 gender */
{
  NON_FEMININE, FEMININE
};

static const int g8tab[] =      /* genders of cardinal numerals,
                                 substantival declension */
{
  MASCULINE_INANIMATE, NEUTER
};

static const int g9tab[] =      /* genders of conjugation, singular */
{
  MASCULINE, FEMININE, NEUTER
};

static const int ptab[] =       /* persons */
{
  PERSON1, PERSON2, PERSON3
};

static const int p1tab[] =      /* persons for simple forms of
                                 imperative in singular */
{
  PERSON2
};


static const int p2tab[] =      /* persons for simple forms of
                                 imperative in plural */
{
  PERSON1, PERSON2
};

static const int dtab[] =       /* degrees */
{
  POSITIVE, COMPARATIVE, SUPERLATIVE
};

static const int atab[] =       /* aspects */
{
  IMPERFECT, PERFECT
};

static const int pmtab[] =      /* punctuation marks */
{
  FULL_STOP, SEMICOLON, COMMA, COLON, DASH, DOTS, QUESTION_MARK,
  EXCLAMATION_MARK, LEFT_ROUND_BRACKET, RIGHT_ROUND_BRACKET,
  LEFT_SQUARE_BRACKET, RIGHT_SQUARE_BRACKET, SIMPLE_BRACKET,
  LEFT_CURLY_BRACKET, RIGHT_CURLY_BRACKET, LEFT_ANGLE_BRACKET,
  RIGHT_ANGLE_BRACKET, LEFT_QUOTATION_MARK, RIGHT_QUOTATION_MARK,
  LEFT_ANGLE_QUOTATION_MARK, RIGHT_ANGLE_QUOTATION_MARK,
  LEFT_DEFINITION_QUOTATION_MARK, RIGHT_DEFINITION_QUOTATION_MARK,
  TYPEWRITER_QUOTATION_MARK
};

static const int gstab[] =      /* graphic signs */
{
  EQUALS_SIGN, PLUS_SIGN, PERCENT_SIGN, PARAGRAPH_SIGN, NUMBER_SIGN,
  DOLLAR_SIGN, AMPERSAND, ASTERISK, COMMERCIAL_AT, BACKSLASH, UP_ARROW,
  UNDERLINE, VERTICAL_LINE, TILDE
};

static const int petab[] =      /* personal endings */
{
  0, PERSON1_SINGULAR, PERSON2_SINGULAR, PERSON1_PLURAL, PERSON2_PLURAL
};

static const int pztab[] =      /* particle ze */
{
  0, ZE_PRESENT
};

/* Symbols used in text output. */
static const char *const types_of_lexemes[] =
{
  "0", "NOUN", "ADJ", "PRON", "NUM", "V", "PASSP", "ACTP", "VN",
  "PASTP", "ADV", "PREP", "CONJ", "PART", "INTERJ", "NONE",
  "UNWORD", "PMARK", "HYPHEN", "NUMBER", "GSIGN", "UNCHAR"
};
static const char *const types_of_pronouns[] =
{
  "0", "pers", "refl", "ind", "neg", "dem", "poss", "intr"
};
static const char *const types_of_numerals[] =
{
  "0", "card", "coll", "frac", "ord"
};
static const char *const forms_of_verbs[] =
{
  "0", "inf", "simp", "past", "imp", "subj", "ifpt", "advp"
};
static const char *const inflexions[] =
{
  "0", "Ps", "Pa", "Pn", "P1", "P2", "Pc", "Pg", "Pj", "Pk", "Pi"
};
static const char *const declension_cases[] =
{
  "0", "nom", "gen", "dat", "acc", "instr", "loc", "voc"
};
static const char *const numbers[] = {"0", "sg", "pl"};
static const char *const genders[] =
{
  "0", "m", "m1", "m2", "m3", "f", "n", "m-3", "m-1", "-m1", "-f"
};
static const char *const persons[] = {"0", "1", "2", "3"};
static const char *const degrees[] = {"0", "pos", "comp", "sup"};
static const char *const aspects[] = {"0", "impf", "pf"};
static const char *const types_of_punctuation_marks[] =
{
  "0", ".", ";", ",", ":", "--", "...", "\?", "!", "(", ")", "[", "]",
  "/", "{", "}", "<", ">", ",,", "\'\'", "<<", ">>", "`", "\'", "\""
};
static const char *const types_of_graphic_signs[] =
{
  "0", "=", "+", "%", "\247", "#", "$", "&", "*", "@", "\\", "^", "_",
  "|", "~"
};
static const char *const personal_endings[] =
{
  "0", "1sg", "2sg", "1pl", "2pl"
};
static const char *const particle_ze[] =
{
  "0", "ze"
};

void lsym_clear(lsym *ls)
{
  ls->s = ls->top = ls->ton = ls->fov = ls->i = ls->c = ls->n = ls->g =
    ls->p = ls->d = ls->a = ls->pm = ls->gs = ls->pe = ls->pz = 0;
}

static int ipresent(int x, const int a[], int n)
{
  int i;
  for (i = 0; i < n; i++)
    if (x == a[i])
      return 1;
  return 0;
}

/*
 * Returns non-zero if ls is a valid lexical symbol. Empty lexical
 * symbol (just after a call to lsym_clear) is invalid.
 */
int lsym_isvalid(const lsym *ls)
{
#define PRESENT(x, a) \
  STMT(if (!ipresent((x), (a), NELEMS(a))) return 0;)
#define ASSERT_NONE(x) STMT(if ((x) != 0) return 0;)

  switch (ls->s)
  {
    case NOUN:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->g, g1tab);
      PRESENT(ls->n, ntab);
      PRESENT(ls->c, ctab);
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case ADJECTIVE:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->d, dtab);
      PRESENT(ls->n, ntab);
      if (ls->n == SINGULAR)
        PRESENT(ls->g, g2tab);
      else
        PRESENT(ls->g, g3tab);
      PRESENT(ls->c, ctab);
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case PRONOUN:
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->i, i1tab);
      switch (ls->i)
      {
        case 0:
          PRESENT(ls->top, top1tab);
          ASSERT_NONE(ls->c);
          ASSERT_NONE(ls->n);
          ASSERT_NONE(ls->g);
          break;
        case SUBSTANTIVAL_DECLENSION:
          PRESENT(ls->top, top2tab);
          PRESENT(ls->n, ntab);
          PRESENT(ls->c, ctab);
          ASSERT_NONE(ls->g);
          break;
        case ADJECTIVAL_DECLENSION:
          PRESENT(ls->top, top3tab);
          PRESENT(ls->n, ntab);
          if (ls->n == SINGULAR)
            PRESENT(ls->g, g2tab);
          else
            PRESENT(ls->g, g3tab);
          PRESENT(ls->c, ctab);
          break;
        case NUMERAL_DECLENSION:
          PRESENT(ls->top, top4tab);
          ASSERT_NONE(ls->n);
          PRESENT(ls->g, g4tab);
          PRESENT(ls->c, ctab);
          break;
        case SINGULAR_DECLENSION_BY_CASE:
          PRESENT(ls->top, top5tab);
          ASSERT_NONE(ls->n);
          ASSERT_NONE(ls->g);
          PRESENT(ls->c, ctab);
          if (ls->top == REFLEXIVE &&
              (ls->c == NOMINATIVE ||
               ls->c == VOCATIVE))
            return 0;
          break;
        default:
          assert(0);
      }
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case NUMERAL:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->ton, tontab);
      switch (ls->ton)
      {
        case CARDINAL:
          PRESENT(ls->i, i2tab);
          switch (ls->i)
          {
            case SUBSTANTIVAL_DECLENSION:
              PRESENT(ls->g, g8tab);
              PRESENT(ls->n, ntab);
              PRESENT(ls->c, ctab);
              break;
            case NUMERAL_DECLENSION:
              ASSERT_NONE(ls->n);
              PRESENT(ls->g, g4tab);
              PRESENT(ls->c, ctab);
              break;
            case DECLENSION_OF_NUMERAL_1:
              ASSERT_NONE(ls->n);
              PRESENT(ls->g, g5tab);
              PRESENT(ls->c, ctab);
              break;
            case DECLENSION_OF_NUMERAL_2:
              ASSERT_NONE(ls->n);
              PRESENT(ls->g, g6tab);
              PRESENT(ls->c, ctab);
              break;
            default:
              assert(0);
          }
          break;
        case COLLECTIVE:
          ASSERT_NONE(ls->i);
          ASSERT_NONE(ls->n);
          ASSERT_NONE(ls->g);
          PRESENT(ls->c, ctab);
          break;
        case FRACTIONAL:
          PRESENT(ls->i, i3tab);
          switch (ls->i)
          {
            case 0:
              ASSERT_NONE(ls->g);
              break;
            case SINGULAR_DECLENSION_BY_GENDER:
              PRESENT(ls->g, g7tab);
              break;
            default:
              assert(0);
          }
          ASSERT_NONE(ls->n);
          ASSERT_NONE(ls->c);
          break;
        case ORDINAL:
          ASSERT_NONE(ls->i);
          PRESENT(ls->n, ntab);
          if (ls->n == SINGULAR)
            PRESENT(ls->g, g2tab);
          else
            PRESENT(ls->g, g3tab);
          PRESENT(ls->c, ctab);
          break;
        default:
          assert(0);
      }
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case VERB:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      ASSERT_NONE(ls->pe);
      PRESENT(ls->fov, fovtab);
      PRESENT(ls->a, atab);
      switch (ls->fov)
      {
        case INFINITIVE:
        case IMPERSONAL_FORM_OF_PAST_TENSE:
        case ADVERBIAL_PARTICIPLE:
          ASSERT_NONE(ls->n);
          ASSERT_NONE(ls->g);
          ASSERT_NONE(ls->p);
          break;
        case SIMPLE_PRESENT_OR_FUTURE:
          ASSERT_NONE(ls->g);
          PRESENT(ls->n, ntab);
          PRESENT(ls->p, ptab);
          break;
        case PAST:
        case SUBJUNCTIVE_MOOD:
          PRESENT(ls->n, ntab);
          if (ls->n == SINGULAR)
            PRESENT(ls->g, g9tab);
          else
            PRESENT(ls->g, g3tab);
          PRESENT(ls->p, ptab);
          break;
        case IMPERATIVE_MOOD:
          ASSERT_NONE(ls->g);
          PRESENT(ls->n, ntab);
          if (ls->n == SINGULAR)
            PRESENT(ls->p, p1tab);
          else
            PRESENT(ls->p, p2tab);
          break;
        default:
          assert(0);
      }
      PRESENT(ls->pz, pztab);
      break;
    case ADJECTIVAL_PASSIVE_PARTICIPLE:
    case ADJECTIVAL_ACTIVE_PARTICIPLE:
    case ADJECTIVAL_PAST_PARTICIPLE:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->n, ntab);
      if (ls->n == SINGULAR)
        PRESENT(ls->g, g2tab);
      else
        PRESENT(ls->g, g3tab);
      PRESENT(ls->c, ctab);
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case VERBAL_NOUN:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->n, ntab);
      PRESENT(ls->c, ctab);
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case ADVERB:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->n);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->d, dtab);
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case PREPOSITION:
    case CONJUNCTION:
    case PARTICLE:
    case INTERJECTION:
    case WORD_OF_NONE_PART_OF_SPEECH:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->n);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      PRESENT(ls->pe, petab);
      PRESENT(ls->pz, pztab);
      break;
    case UNKNOWN_WORD:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->n);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      ASSERT_NONE(ls->pe);
      ASSERT_NONE(ls->pz);
      break;
    case PUNCTUATION_MARK:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->n);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      PRESENT(ls->pm, pmtab);
      ASSERT_NONE(ls->gs);
      ASSERT_NONE(ls->pe);
      ASSERT_NONE(ls->pz);
      break;
    case GRAPHIC_SIGN:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->n);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      PRESENT(ls->gs, gstab);
      ASSERT_NONE(ls->pe);
      ASSERT_NONE(ls->pz);
      break;
    case HYPHEN:
    case NUMBER:
    case UNKNOWN_CHARACTER:
      ASSERT_NONE(ls->top);
      ASSERT_NONE(ls->ton);
      ASSERT_NONE(ls->fov);
      ASSERT_NONE(ls->i);
      ASSERT_NONE(ls->c);
      ASSERT_NONE(ls->n);
      ASSERT_NONE(ls->g);
      ASSERT_NONE(ls->p);
      ASSERT_NONE(ls->d);
      ASSERT_NONE(ls->a);
      ASSERT_NONE(ls->pm);
      ASSERT_NONE(ls->gs);
      ASSERT_NONE(ls->pe);
      ASSERT_NONE(ls->pz);
      break;
    default:
      return 0;
  }
  return 1;

#undef PRESENT
#undef ASSERT_NONE
}

/* True if two lexical symbols are equal. */
int lsym_eq(const lsym *ls1, const lsym *ls2)
{
  return
    ls1->s == ls2->s &&
    ls1->top == ls2->top &&
    ls1->ton == ls2->ton &&
    ls1->fov == ls2->fov &&
    ls1->i == ls2->i &&
    ls1->c == ls2->c &&
    ls1->n == ls2->n &&
    ls1->g == ls2->g &&
    ls1->p == ls2->p &&
    ls1->d == ls2->d &&
    ls1->a == ls2->a &&
    ls1->pm == ls2->pm &&
    ls1->gs == ls2->gs &&
    ls1->pe == ls2->pe &&
    ls1->pz == ls2->pz;
}

static void bstrcat(char *s, const char *t)
{
  strcat(strcat(s, "("), t);
}

static void cstrcat(char *s, const char *t)
{
  strcat(strcat(s, ","), t);
}

static void fstrcat(char *s, const lsym *ls, int open)
{
  if (ls->pe != 0)
  {
    if (open)
    {
      bstrcat(s, personal_endings[ls->pe]);
      open = 0;
    }
    else
      cstrcat(s, personal_endings[ls->pe]);
  }
  if (ls->pz != 0)
  {
    if (open)
    {
      bstrcat(s, particle_ze[ls->pz]);
      open = 0;
    }
    else
      cstrcat(s, particle_ze[ls->pz]);
  }
  if (!open)
    strcat(s, ")");
}

/*
 * Returns character representation of a lexical symbol ls. The symbol
 * ls should be a valid symbol (assert(lsym_isvalid(ls))), if the symbol
 * is invalid, the result is undefined. The string s must be at least
 * MAXSYMLEN characters long.
 */
char *lsym_tostr(const lsym *ls, char s[])
{

  strcpy(s, types_of_lexemes[ls->s]);

  switch (ls->s)
  {
    case NOUN:
      bstrcat(s, genders[ls->g]);
      cstrcat(s, numbers[ls->n]);
      cstrcat(s, declension_cases[ls->c]);
      fstrcat(s, ls, 0);
      break;
    case ADJECTIVE:
      bstrcat(s, degrees[ls->d]);
      cstrcat(s, numbers[ls->n]);
      cstrcat(s, genders[ls->g]);
      cstrcat(s, declension_cases[ls->c]);
      fstrcat(s, ls, 0);
      break;
    case PRONOUN:
      bstrcat(s, types_of_pronouns[ls->top]);
      cstrcat(s, inflexions[ls->i]);
      switch (ls->i)
      {
        case 0:
          break;
        case SUBSTANTIVAL_DECLENSION:
          cstrcat(s, numbers[ls->n]);
          cstrcat(s, declension_cases[ls->c]);
          break;
        case ADJECTIVAL_DECLENSION:
          cstrcat(s, numbers[ls->n]);
          cstrcat(s, genders[ls->g]);
          cstrcat(s, declension_cases[ls->c]);
          break;
        case NUMERAL_DECLENSION:
          cstrcat(s, genders[ls->g]);
          cstrcat(s, declension_cases[ls->c]);
          break;
        case SINGULAR_DECLENSION_BY_CASE:
          cstrcat(s, declension_cases[ls->c]);
          break;
      }
      fstrcat(s, ls, 0);
      break;
    case NUMERAL:
      bstrcat(s, types_of_numerals[ls->ton]);
      switch (ls->ton)
      {
        case CARDINAL:
          cstrcat(s, inflexions[ls->i]);
          switch (ls->i)
          {
            case SUBSTANTIVAL_DECLENSION:
              cstrcat(s, genders[ls->g]);
              cstrcat(s, numbers[ls->n]);
              cstrcat(s, declension_cases[ls->c]);
              break;
            case NUMERAL_DECLENSION:
            case DECLENSION_OF_NUMERAL_1:
            case DECLENSION_OF_NUMERAL_2:
              cstrcat(s, genders[ls->g]);
              cstrcat(s, declension_cases[ls->c]);
              break;
          }
          break;
        case COLLECTIVE:
          cstrcat(s, declension_cases[ls->c]);
          break;
        case FRACTIONAL:
          cstrcat(s, genders[ls->g]);
          break;
        case ORDINAL:
          cstrcat(s, numbers[ls->n]);
          cstrcat(s, genders[ls->g]);
          cstrcat(s, declension_cases[ls->c]);
          break;
      }
      fstrcat(s, ls, 0);
      break;
    case VERB:
      bstrcat(s, aspects[ls->a]);
      cstrcat(s, forms_of_verbs[ls->fov]);
      switch (ls->fov)
      {
        case INFINITIVE:
        case IMPERSONAL_FORM_OF_PAST_TENSE:
        case ADVERBIAL_PARTICIPLE:
          break;
        case SIMPLE_PRESENT_OR_FUTURE:
        case IMPERATIVE_MOOD:
          cstrcat(s, numbers[ls->n]);
          cstrcat(s, persons[ls->p]);
          break;
        case PAST:
        case SUBJUNCTIVE_MOOD:
          cstrcat(s, numbers[ls->n]);
          cstrcat(s, genders[ls->g]);
          cstrcat(s, persons[ls->p]);
          break;
      }
      fstrcat(s, ls, 0);
      break;
    case ADJECTIVAL_PASSIVE_PARTICIPLE:
    case ADJECTIVAL_ACTIVE_PARTICIPLE:
    case ADJECTIVAL_PAST_PARTICIPLE:
      bstrcat(s, numbers[ls->n]);
      cstrcat(s, genders[ls->g]);
      cstrcat(s, declension_cases[ls->c]);
      fstrcat(s, ls, 0);
      break;
    case VERBAL_NOUN:
      bstrcat(s, numbers[ls->n]);
      cstrcat(s, declension_cases[ls->c]);
      fstrcat(s, ls, 0);
      break;
    case ADVERB:
      bstrcat(s, degrees[ls->d]);
      fstrcat(s, ls, 0);
      break;
    case PREPOSITION:
    case CONJUNCTION:
    case PARTICLE:
    case INTERJECTION:
    case WORD_OF_NONE_PART_OF_SPEECH:
      fstrcat(s, ls, 1);
      break;
    case UNKNOWN_WORD:
    case HYPHEN:
    case NUMBER:
    case UNKNOWN_CHARACTER:
      break;
    case PUNCTUATION_MARK:
      bstrcat(s, types_of_punctuation_marks[ls->pm]);
      strcat(s, ")");
      break;
    case GRAPHIC_SIGN:
      bstrcat(s, types_of_graphic_signs[ls->gs]);
      strcat(s, ")");
      break;
  }
  return s;
}

static void output(const lsym *ls, FILE *f)
{
  char s[MAXSYMLEN + 1];
  assert(lsym_isvalid(ls));
  fprintf(f, "%s\n", lsym_tostr(ls, s));
}

/* Generates inflexion by case, 7 forms. */
static void cgen(lsym *ls, FILE *f)
{
  int c;

  for (c = 0; c < NELEMS(ctab); c++)
  {
    ls->c = ctab[c];
    output(ls, f);
  }
}

/* Generates substantival declension, 14 forms. */
static void sgen(lsym *ls, FILE *f)
{
  ls->n = SINGULAR;
  cgen(ls, f);
  ls->n = PLURAL;
  cgen(ls, f);
}

/* Generates adjectival declension, 35 forms. */
static void agen(lsym *ls, FILE *f)
{
  int g;

  ls->n = SINGULAR;
  for (g = 0; g < NELEMS(g2tab); g++)
  {
    ls->g = g2tab[g];
    cgen(ls, f);
  }
  ls->n = PLURAL;
  for (g = 0; g < NELEMS(g3tab); g++)
  {
    ls->g = g3tab[g];
    cgen(ls, f);
  }
}

/* Generates inflexion by person, 3 forms. */
static void pgen(lsym *ls, FILE *f)
{
  int p;

  for (p = 0; p < NELEMS(ptab); p++)
  {
    ls->p = ptab[p];
    output(ls, f);
  }
}

/* Generates conjugation, 30 forms. */
static void jgen(lsym *ls, FILE *f)
{
  int g;

  ls->n = SINGULAR;
  for (g = 0; g < NELEMS(g9tab); g++)
  {
    ls->g = g9tab[g];
    pgen(ls, f);
  }
  ls->n = PLURAL;
  for (g = 0; g < NELEMS(g3tab); g++)
  {
    ls->g = g3tab[g];
    pgen(ls, f);
  }
}

static void generate(const lsym *ls0, lsym *ls, FILE *f)
{
  int t, i, c, n, g, p, d, a;

  *ls = *ls0; ls->s = NOUN;
  for (g = 0; g < NELEMS(g1tab); g++)
  {
    ls->g = g1tab[g];
    sgen(ls, f);
  }

  *ls = *ls0; ls->s = ADJECTIVE;
  for (d = 0; d < NELEMS(dtab); d++)
  {
    ls->d = dtab[d];
    agen(ls, f);
  }

  for (i = 0; i < NELEMS(i1tab); i++)
  {
    *ls = *ls0; ls->s = PRONOUN;
    ls->i = i1tab[i];
    switch (ls->i)
    {
      case 0:
        for (t = 0; t < NELEMS(top1tab); t++)
        {
          ls->top = top1tab[t];
          output(ls, f);
        }
        break;
      case SUBSTANTIVAL_DECLENSION:
        for (t = 0; t < NELEMS(top2tab); t++)
        {
          ls->top = top2tab[t];
          sgen(ls, f);
        }
        break;
      case ADJECTIVAL_DECLENSION:
        for (t = 0; t < NELEMS(top3tab); t++)
        {
          ls->top = top3tab[t];
          agen(ls, f);
        }
        break;
      case NUMERAL_DECLENSION:
        for (t = 0; t < NELEMS(top4tab); t++)
        {
          ls->top = top4tab[t];
          for (g = 0; g < NELEMS(g4tab); g++)
          {
            ls->g = g4tab[g];
            cgen(ls, f);
          }
        }
        break;
      case SINGULAR_DECLENSION_BY_CASE:
        for (t = 0; t < NELEMS(top5tab); t++)
        {
          ls->top = top5tab[t];
          for (c = 0; c < NELEMS(ctab); c++)
          {
            ls->c = ctab[c];
            if (ls->top == REFLEXIVE &&
                (ls->c == NOMINATIVE ||
                 ls->c == VOCATIVE))
              continue;
            output(ls, f);
          }
        }
        break;
    }
  }

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = CARDINAL;
  ls->i = SUBSTANTIVAL_DECLENSION;
  for (g = 0; g < NELEMS(g8tab); g++)
  {
    ls->g = g8tab[g];
    sgen(ls, f);
  }

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = CARDINAL;
  ls->i = NUMERAL_DECLENSION;
  for (g = 0; g < NELEMS(g4tab); g++)
  {
    ls->g = g4tab[g];
    cgen(ls, f);
  }

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = CARDINAL;
  ls->i = DECLENSION_OF_NUMERAL_1;
  for (g = 0; g < NELEMS(g5tab); g++)
  {
    ls->g = g5tab[g];
    cgen(ls, f);
  }

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = CARDINAL;
  ls->i = DECLENSION_OF_NUMERAL_2;
  for (g = 0; g < NELEMS(g6tab); g++)
  {
    ls->g = g6tab[g];
    cgen(ls, f);
  }

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = COLLECTIVE;
  cgen(ls, f);

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = FRACTIONAL;
  output(ls, f);
  ls->i = SINGULAR_DECLENSION_BY_GENDER;
  for (g = 0; g < NELEMS(g7tab); g++)
  {
    ls->g = g7tab[g];
    output(ls, f);
  }

  *ls = *ls0; ls->s = NUMERAL;
  ls->ton = ORDINAL;
  agen(ls, f);

  if (ls0->pe == 0)
    for (a = 0; a < NELEMS(atab); a++)
    {
      *ls = *ls0; ls->s = VERB;
      ls->a = atab[a];
      ls->fov = INFINITIVE;
      output(ls, f);

      ls->fov = SIMPLE_PRESENT_OR_FUTURE;
      for (n = 0; n < NELEMS(ntab); n++)
      {
        ls->n = ntab[n];
        pgen(ls, f);
      }

      *ls = *ls0; ls->s = VERB;
      ls->a = atab[a];
      ls->fov = PAST;
      jgen(ls, f);

      *ls = *ls0; ls->s = VERB;
      ls->a = atab[a];
      ls->fov = IMPERATIVE_MOOD;
      ls->n = SINGULAR;
      for (p = 0; p < NELEMS(p1tab); p++)
      {
        ls->p = p1tab[p];
        output(ls, f);
      }
      ls->n = PLURAL;
      for (p = 0; p < NELEMS(p2tab); p++)
      {
        ls->p = p2tab[p];
        output(ls, f);
      }

      *ls = *ls0; ls->s = VERB;
      ls->a = atab[a];
      ls->fov = SUBJUNCTIVE_MOOD;
      jgen(ls, f);

      *ls = *ls0; ls->s = VERB;
      ls->a = atab[a];
      ls->fov = IMPERSONAL_FORM_OF_PAST_TENSE;
      output(ls, f);

      *ls = *ls0; ls->s = VERB;
      ls->a = atab[a];
      ls->fov = ADVERBIAL_PARTICIPLE;
      output(ls, f);
    }

  *ls = *ls0; ls->s = ADJECTIVAL_PASSIVE_PARTICIPLE; agen(ls, f);
  *ls = *ls0; ls->s = ADJECTIVAL_ACTIVE_PARTICIPLE;  agen(ls, f);
  *ls = *ls0; ls->s = VERBAL_NOUN;                   sgen(ls, f);
  *ls = *ls0; ls->s = ADJECTIVAL_PAST_PARTICIPLE;    agen(ls, f);

  *ls = *ls0; ls->s = ADVERB;
  for (d = 0; d < NELEMS(dtab); d++)
  {
    ls->d = dtab[d];
    output(ls, f);
  }

  *ls = *ls0; ls->s = PREPOSITION;                 output(ls, f);
  *ls = *ls0; ls->s = CONJUNCTION;                 output(ls, f);
  *ls = *ls0; ls->s = PARTICLE;                    output(ls, f);
  *ls = *ls0; ls->s = INTERJECTION;                output(ls, f);
  *ls = *ls0; ls->s = WORD_OF_NONE_PART_OF_SPEECH; output(ls, f);

  if (ls0->pe == 0 && ls0->pz == 0)
  {
    *ls = *ls0; ls->s = UNKNOWN_WORD;
    output(ls, f);
    *ls = *ls0; ls->s = PUNCTUATION_MARK;
    for (p = 0; p < NELEMS(pmtab); p++)
    {
      ls->pm = pmtab[p];
      output(ls, f);
    }
    *ls = *ls0; ls->s = GRAPHIC_SIGN;
    for (p = 0; p < NELEMS(gstab); p++)
    {
      ls->gs = gstab[p];
      output(ls, f);
    }
    *ls = *ls0; ls->s = HYPHEN;            output(ls, f);
    *ls = *ls0; ls->s = NUMBER;            output(ls, f);
    *ls = *ls0; ls->s = UNKNOWN_CHARACTER; output(ls, f);
  }
}

/* Generates list of all lexical symbols to the file f. */
void lsym_genlist(FILE *f)
{
  lsym ls0, ls;
  int i, j;

  lsym_clear(&ls0);
  for (i = 0; i < NELEMS(petab); i++)
  {
    ls0.pe = petab[i];
    for (j = 0; j < NELEMS(pztab); j++)
    {
      ls0.pz = pztab[j];
      generate(&ls0, &ls, f);
    }
  }
}

/*
 * The function generates a superset of lexical symbols. Punctuation
 * marks and graphic signs are excluded from this test. If a symbol is
 * valid, it should be printed. At the end, the result should be the
 * same as the result of lsym_genlist, but with no punctuation marks and
 * graphic signs. The test takes about 16 minutes on an 686 processor
 * with 1 GHz.
 */
void lsym_ltest(void)
{
  lsym ls;
  char s[MAXSYMLEN + 1];

  lsym_clear(&ls);

  for (ls.s = 0;
       ls.s <= 1 + UNKNOWN_CHARACTER;
       ls.s++)
    for (ls.top = 0;
         ls.top <= 1 + INTERROGATIVE_RELATIVE;
         ls.top++)
      for (ls.ton = 0;
           ls.ton <= 1 + ORDINAL;
           ls.ton++)
        for (ls.fov = 0;
             ls.fov <= 1 + ADVERBIAL_PARTICIPLE;
             ls.fov++)
          for (ls.i = 0;
               ls.i <= 1 + SIMPLE_FORMS_OF_IMPERATIVE;
               ls.i++)
            for (ls.c = 0;
                 ls.c <= 1 + VOCATIVE;
                 ls.c++)
              for (ls.n = 0;
                   ls.n <= 1 + PLURAL;
                   ls.n++)
                for (ls.g = 0;
                     ls.g <= 1 + NON_FEMININE;
                     ls.g++)
                  for (ls.p = 0;
                       ls.p <= 1 + PERSON3;
                       ls.p++)
                    for (ls.d = 0;
                         ls.d <= 1 + SUPERLATIVE;
                         ls.d++)
                      for (ls.a = 0;
                           ls.a <= 1 + PERFECT;
                           ls.a++)
                        for (ls.pe = 0;
                             ls.pe <= 1 + PERSON2_PLURAL;
                             ls.pe++)
                          for (ls.pz = 0;
                               ls.pz <= 1 + ZE_PRESENT;
                               ls.pz++)
                          {
                            if (lsym_isvalid(&ls))
                              printf("%s\n", lsym_tostr(&ls, s));
                          }
}

/* Prints lexical symbol for debugging. */
void lsym_print(const lsym *ls)
{
  printf("symbol:           %d\n", ls->s);
  printf("type of pronoun:  %d\n", ls->top);
  printf("type of numeral:  %d\n", ls->ton);
  printf("form of verb:     %d\n", ls->fov);
  printf("inflexion:        %d\n", ls->i);
  printf("case:             %d\n", ls->c);
  printf("number:           %d\n", ls->n);
  printf("gender:           %d\n", ls->g);
  printf("person:           %d\n", ls->p);
  printf("degree:           %d\n", ls->d);
  printf("aspect:           %d\n", ls->a);
  printf("punctuation mark: %d\n", ls->pm);
  printf("graphic sign:     %d\n", ls->gs);
  printf("personal ending:  %d\n", ls->pe);
  printf("particle ze:      %d\n", ls->pz);
}

/**************************************************************
 *                                                            *
 *                        Dictionary.                         *
 *                                                            *
 **************************************************************/

#define NINFLS 8                /* number of inflexions */
#define NTYPES 47               /* number of types of entries */
#define MAXFORMS 42             /* maximum of nforms */

struct dictsize                 /* size of a dictionary */
{
  int nentries;                 /* number of entries */
  int nendings[NINFLS];         /* numbers of endings */
  int stembuflen;               /* length of buffer for stems */
  int endbuflen;                /* length of buffer for endings */
};

struct entry                    /* dictionary entry */
{
  char type;                    /* type of entry (see table head) */
  char *stem;                   /* its stem */
  short ending;                 /* index of its ending */
};

/*
 * Dictionary, as it resides in memory ready to work.
 *
 * For 0 <= i < maxsize.nentries, the i-th entry can be found as
 * enttab[i]. The number t = enttab[i].type is a type of this entry, 0
 * <= t < ntypes, and f = infl[t] is an inflection corresponding to this
 * type. If f == -1, the entry is uninflected and its only form is the
 * stem enttab[i].stem; in this case, enttab[i].ending = -1. If f >= 0,
 * the entry is inflected and the number of forms of the entry is nf =
 * nforms[f]. Each form can be constructed by concatenating the stem
 * enttab[i].stem and one of the endings endtab[f][enttab[i].ending][j],
 * 0 <= j < nf. If a particular form is missing, it is marked with the
 * j-th ending by setting it to noform.
 *
 * A similar construction is described in Z. Vetulani, J. Martinek, T.
 * Obrebski, G. Vetulani, Dictionary Based Methods and Tools for
 * Language Engineering, Wydawnictwo Naukowe UAM, Poznan 1998, chapter
 * 3.
 */
struct dict
{
  struct dictsize maxsize;      /* maximum size of the dictionary */
  struct entry *enttab;         /* table of entries */
  char ***endtab[NINFLS];       /* table of endings */
  char *stembuf;                /* buffer for stems */
  char *endbuf;                 /* buffer for endings */
};

static const char wfmark[] = "PLA word file";

/* size of dictionary used in dict_compile */
static const struct dictsize maxsz =
{
  140000, {1300, 70, 40, 1, 5, 16, 1, 500}, 2000000, 2000000
};

/* missing form of inflected entry */
static const char *const noform = "-";

/* numbers of forms in inflexions */
static const int nforms[NINFLS] =
{
  14,                           /* 0 - substantival declension */
  42,                           /* 1 - adjectival declension */
  14,                           /* 2 - numeral declension */
  28,                           /* 3 - declension of numeral 1 */
  28,                           /* 4 - declension of numeral 2 */
  7,                            /* 5 - declension by case */
  2,                            /* 6 - inflexion by gender */
  42                            /* 7 - conjugation */
};

/*
 * Inflexions corresponding to types of entries. -1 means uninflected,
 * others as in nforms. This table corresponds to the table head: entry
 * of type t named head[t] is inflected according to inflexion infl[t]
 * and if infl[t] > 0, the inflexion possesses nforms[infl[t]] forms,
 * otherwise the entry is uninflected and possesses one form.
 */
static const int infl[NTYPES] =
{
  0,  0,  0,  0,  0,  0,  1,  1,  0,  1,  5,  5,  1,  2, -1,  5,
  1, -1,  1,  2, -1,  1,  5,  1,  2, -1,  0,  2,  3,  4,  5, -1,
  6,  1,  7,  7,  1,  1,  0,  1, -1, -1, -1, -1, -1, -1, -1
};

/*
 * Headings of types of entries as they appear in source files for
 * dictionaries.
 */
static const char *const head[NTYPES] =
{
  "MASCULINE-PERSONAL NOUN",                               /*  0 */
  "MASCULINE-ANIMAL NOUN",                                 /*  1 */
  "MASCULINE-INANIMATE NOUN",                              /*  2 */
  "FEMININE NOUN",                                         /*  3 */
  "NEUTER NOUN",                                           /*  4 */
  "GENDERLESS NOUN",                                       /*  5 */
  "POSITIVE ADJECTIVE",                                    /*  6 */
  "COMPARATIVE ADJECTIVE",                                 /*  7 */
  "PERSONAL PRONOUN, SUBSTANTIVAL DECLENSION",             /*  8 */
  "PERSONAL PRONOUN, ADJECTIVAL DECLENSION",               /*  9 */
  "REFLEXIVE PRONOUN, DECLENSION BY CASE",                 /* 10 */
  "INDEFINITE PRONOUN, DECLENSION BY CASE",                /* 11 */
  "INDEFINITE PRONOUN, ADJECTIVAL DECLENSION",             /* 12 */
  "INDEFINITE PRONOUN, NUMERAL DECLENSION",                /* 13 */
  "INDEFINITE PRONOUN, UNINFLECTED",                       /* 14 */
  "NEGATIVE PRONOUN, DECLENSION BY CASE",                  /* 15 */
  "NEGATIVE PRONOUN, ADJECTIVAL DECLENSION",               /* 16 */
  "NEGATIVE PRONOUN, UNINFLECTED",                         /* 17 */
  "DEMONSTRATIVE PRONOUN, ADJECTIVAL DECLENSION",          /* 18 */
  "DEMONSTRATIVE PRONOUN, NUMERAL DECLENSION",             /* 19 */
  "DEMONSTRATIVE PRONOUN, UNINFLECTED",                    /* 20 */
  "POSSESSIVE PRONOUN, ADJECTIVAL DECLENSION",             /* 21 */
  "INTERROGATIVE-RELATIVE PRONOUN, DECLENSION BY CASE",    /* 22 */
  "INTERROGATIVE-RELATIVE PRONOUN, ADJECTIVAL DECLENSION", /* 23 */
  "INTERROGATIVE-RELATIVE PRONOUN, NUMERAL DECLENSION",    /* 24 */
  "INTERROGATIVE-RELATIVE PRONOUN, UNINFLECTED",           /* 25 */
  "CARDINAL NUMERAL, SUBSTANTIVAL DECLENSION",             /* 26 */
  "CARDINAL NUMERAL, NUMERAL DECLENSION",                  /* 27 */
  "CARDINAL NUMERAL, DECLENSION OF NUMERAL 1",             /* 28 */
  "CARDINAL NUMERAL, DECLENSION OF NUMERAL 2",             /* 29 */
  "COLLECTIVE NUMERAL, DECLENSION BY CASE",                /* 30 */
  "FRACTIONAL NUMERAL, UNINFLECTED",                       /* 31 */
  "FRACTIONAL NUMERAL, INFLECTED BY GENDER",               /* 32 */
  "ORDINAL NUMERAL, ADJECTIVAL DECLENSION",                /* 33 */
  "IMPERFECT VERB",                                        /* 34 */
  "PERFECT VERB",                                          /* 35 */
  "ADJECTIVAL PASSIVE PARTICIPLE",                         /* 36 */
  "ADJECTIVAL ACTIVE PARTICIPLE",                          /* 37 */
  "VERBAL NOUN",                                           /* 38 */
  "ADJECTIVAL PAST PARTICIPLE",                            /* 39 */
  "POSITIVE ADVERB",                                       /* 40 */
  "COMPARATIVE ADVERB",                                    /* 41 */
  "PREPOSITION",                                           /* 42 */
  "CONJUNCTION",                                           /* 43 */
  "PARTICLE",                                              /* 44 */
  "INTERJECTION",                                          /* 45 */
  "WORD OF NONE PART OF SPEECH"                            /* 46 */
};

/* Lexical information which can be deduced from type of entry. */
static const lsym lexical_info[NTYPES] =
{
  {NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0,
   MASCULINE_PERSONAL, 0, 0, 0, 0, 0, 0, 0},
  {NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0,
   MASCULINE_ANIMAL, 0, 0, 0, 0, 0, 0, 0},
  {NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0,
   MASCULINE_INANIMATE, 0, 0, 0, 0, 0, 0, 0},
  {NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0,
   FEMININE, 0, 0, 0, 0, 0, 0, 0},
  {NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0,
   NEUTER, 0, 0, 0, 0, 0, 0, 0},
  {NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0},
  {ADJECTIVE, 0, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0,
   POSITIVE, 0, 0, 0, 0, 0},
  {ADJECTIVE, 0, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0,
   COMPARATIVE, 0, 0, 0, 0, 0},
  {PRONOUN,
   PERSONAL, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   PERSONAL, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   REFLEXIVE, 0, 0,
   SINGULAR_DECLENSION_BY_CASE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INDEFINITE, 0, 0,
   SINGULAR_DECLENSION_BY_CASE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INDEFINITE, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INDEFINITE, 0, 0,
   NUMERAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INDEFINITE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   NEGATIVE, 0, 0,
   SINGULAR_DECLENSION_BY_CASE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   NEGATIVE, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   NEGATIVE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   DEMONSTRATIVE, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   DEMONSTRATIVE, 0, 0,
   NUMERAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   DEMONSTRATIVE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   POSSESSIVE, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INTERROGATIVE_RELATIVE, 0, 0,
   SINGULAR_DECLENSION_BY_CASE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INTERROGATIVE_RELATIVE, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INTERROGATIVE_RELATIVE, 0, 0,
   NUMERAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PRONOUN,
   INTERROGATIVE_RELATIVE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0},
  {NUMERAL, 0,
   CARDINAL, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {NUMERAL, 0,
   CARDINAL, 0,
   NUMERAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {NUMERAL, 0,
   CARDINAL, 0,
   DECLENSION_OF_NUMERAL_1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {NUMERAL, 0,
   CARDINAL, 0,
   DECLENSION_OF_NUMERAL_2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {NUMERAL, 0,
   COLLECTIVE, 0,
   SINGULAR_DECLENSION_BY_CASE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {NUMERAL, 0,
   FRACTIONAL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {NUMERAL, 0,
   FRACTIONAL, 0,
   SINGULAR_DECLENSION_BY_GENDER, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0},
  {NUMERAL, 0,
   ORDINAL, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {VERB, 0, 0, 0,
   CONJUGATION, 0, 0, 0, 0, 0,
   IMPERFECT, 0, 0, 0, 0},
  {VERB, 0, 0, 0,
   CONJUGATION, 0, 0, 0, 0, 0,
   PERFECT, 0, 0, 0, 0},
  {ADJECTIVAL_PASSIVE_PARTICIPLE, 0, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {ADJECTIVAL_ACTIVE_PARTICIPLE, 0, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {VERBAL_NOUN, 0, 0, 0,
   SUBSTANTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {ADJECTIVAL_PAST_PARTICIPLE, 0, 0, 0,
   ADJECTIVAL_DECLENSION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {ADVERB, 0, 0, 0, 0, 0, 0, 0, 0,
   POSITIVE, 0, 0, 0, 0, 0},
  {ADVERB, 0, 0, 0, 0, 0, 0, 0, 0,
   COMPARATIVE, 0, 0, 0, 0, 0},
  {PREPOSITION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {CONJUNCTION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {PARTICLE, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {INTERJECTION, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  {WORD_OF_NONE_PART_OF_SPEECH, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0}
};

void dict_free(dict *d)
{
  if (d != NULL)
  {
    int i, j;
    free(d->endbuf);
    free(d->stembuf);
    i = NINFLS;
    while (i-- > 0)
      if (d->endtab[i] != NULL)
      {
        j = d->maxsize.nendings[i];
        while (j-- > 0)
          free(d->endtab[i][j]);
        free(d->endtab[i]);
      }
    free(d->enttab);
    free(d);
  }
}

static dict *dict_alloc(const struct dictsize *sz)
{
  dict *d;
  int i, j;

  if ((d = NEW(dict)) == NULL)
    goto fin;
  d->maxsize = *sz;
  for (i = 0; i < NINFLS; i++)
    d->endtab[i] = NULL;
  d->stembuf = NULL;
  d->endbuf = NULL;
  if ((d->enttab = NEW_VECTOR(struct entry, sz->nentries)) == NULL)
    goto fin;
  for (i = 0; i < NINFLS; i++)
    if (sz->nendings[i] > 0)
    {
      if ((d->endtab[i] = NEW_VECTOR(char **, sz->nendings[i])) == NULL)
        goto fin;
      for (j = 0; j < sz->nendings[i]; j++)
        d->endtab[i][j] = NULL;
      for (j = 0; j < sz->nendings[i]; j++)
        if ((d->endtab[i][j] = NEW_VECTOR(char *, nforms[i])) == NULL)
          goto fin;
    }
  if (sz->stembuflen > 0)
    if ((d->stembuf = NEW_VECTOR(char, sz->stembuflen)) == NULL)
      goto fin;
  if (sz->endbuflen > 0)
    if ((d->endbuf = NEW_VECTOR(char, sz->endbuflen)) == NULL)
      goto fin;

  return d;

fin:
  dict_free(d);
  return NULL;
}

/* Throw comment out from s. */
static char *throwc(char *s)
{
  char *t;
  for (t = s; *t; t++)
    if (*t == '#')
    {
      *t = '\0';
      return s;
    }
  return s;
}

static int get_entry(FILE *f, int *type,
                     char forms[MAXFORMS][MAXLEN + 1], int *line)
{
  char s[MAXLEN + 1];
  int fl, nf, i, ex;

  /* Skip empty lines and get the heading. */
  for (;;)
  {
    (*line)++;
    if (fgets(s, MAXLEN + 1, f) == NULL)
    {
      *type = -1;
      return 0;
    }
    if (strlen(s) == MAXLEN && s[MAXLEN - 1] != '\n')
      return ERR_BADSWF;        /* line too long */
    if (*trim(throwc(s)))
      break;
  }

  /* Determine type and the number of forms basing on the heading. */
  for (*type = 0; *type < NTYPES; (*type)++)
    if (strcmp(s, head[*type]) == 0)
      break;
  if (*type >= NTYPES)
    return ERR_BADSWF;          /* unknown heading */
  fl = infl[*type];
  nf = fl == -1 ? 1 : nforms[fl];

  /* Read all forms into the table. */
  ex = 0;
  for (i = 0; i < nf; i++)
  {
    (*line)++;
    if (fgets(s, MAXLEN + 1, f) == NULL)
      return ERR_BADSWF;        /* unexpected end of entry */
    if (strlen(s) == MAXLEN && s[MAXLEN - 1] != '\n')
      return ERR_BADSWF;        /* line too long */
    if (!*trim(throwc(s)))
      return ERR_BADSWF;        /* unexpected end of entry */
    if (strcmp(s, noform))
    {
      strcpy(forms[i], s);
      ex = 1;
    }
    else
      *forms[i] = '\0';
  }
  if (!ex)
    return ERR_BADSWF;          /* no form given */

  /* Each entry must be followed by at least one empty line or the end
   of file. */
  (*line)++;
  if (fgets(s, MAXLEN + 1, f) == NULL)
    return 0;
  if (strlen(s) == MAXLEN && s[MAXLEN - 1] != '\n')
    return ERR_BADSWF;          /* line too long */
  if (*trim(s))
    return ERR_BADSWF;          /* end of entry expected */
  return 0;
}

static int add_entry(const int type, char forms[MAXFORMS][MAXLEN + 1],
                     dict *d, struct dictsize *sz)
{
  const int fl = infl[type];
  const int nf = fl == -1 ? 1 : nforms[fl];
  char *stem = NULL;
  int ending = -1;
  int ip, lp;                   /* index and length of the prefix */
  int len = 0;                  /* length of the stem + 1 */
  int i, j;
  char c;

  if (sz->nentries >= d->maxsize.nentries)
    return ERR_TOOBIG;          /* too many entries */

  if (nf == 1)                  /* Entries possessing only one form. */
  {
    /*
     * The following searching for already existing stems seriously
     * extends elapsed time. If it is unacceptable, remove the following
     * loop. The same concerns analogic loop for inflected entries. The
     * loops save about 1/4 of stembufsize. Perhaps it is better to
     * perform this in dict_sort?
     */

    for (i = 0; i < sz->nentries; i++)
      if (strcmp(forms[0], d->enttab[i].stem) == 0)
      {
        stem = d->enttab[i].stem;
        break;
      }
    if (stem == NULL)
    {
      len = strlen(forms[0]) + 1;
      if (sz->stembuflen + len > d->maxsize.stembuflen)
        return ERR_TOOBIG;      /* stem buffer too small */
      stem = strcpy(d->stembuf + sz->stembuflen, forms[0]);
    }
  }
  else                          /* Inflected entries. */
  {
    /*
     * Find the longest common prefix of all non-empty forms. The prefix
     * will be in forms[ip] and will be lp characters long. The
     * existence of the non-empty form is guaranteed by the function
     * get_entry (the variable ex).
     */

    for (ip = 0; !*forms[ip]; ip++);
    lp = strlen(forms[ip]);
    assert(lp > 0);
    for (i = ip + 1; i < nf; i++)
      if (*forms[i])
      {
        for (j = 0; j < lp && forms[i][j] == forms[ip][j]; j++);
        if (j < lp)
          lp = j;
      }

    /* Elaborate the stem. */
    c = forms[ip][lp];
    forms[ip][lp] = 0;
    for (i = 0; i < sz->nentries; i++)
      if (strcmp(forms[ip], d->enttab[i].stem) == 0)
      {
        stem = d->enttab[i].stem;
        break;
      }
    if (stem == NULL)
    {
      len = lp + 1;
      if (sz->stembuflen + len > d->maxsize.stembuflen)
        return ERR_TOOBIG;      /* stem buffer too small */
      stem = strcpy(d->stembuf + sz->stembuflen, forms[ip]);
    }
    forms[ip][lp] = c;

    /* Remove the prefix from non-empty forms to get endings. */
    for (i = 0; i < nf; i++)
      if (*forms[i])
        for (j = lp; (forms[i][j - lp] = forms[i][j]); j++);
      else
        strcpy(forms[i], noform);

    /* Search the table of endings for the set of endings in forms. */
    for (i = 0; i < sz->nendings[fl]; i++)
    {
      for (j = 0; j < nf; j++)
        if (strcmp(forms[j], d->endtab[fl][i][j]))
          break;
      if (j == nf)
      {
        ending = i;
        break;
      }
    }

    /* If the set of endings is not found, try to add it. */
    if (ending == -1)
    {
      int z = sz->endbuflen;
      if ((ending = sz->nendings[fl]) >= d->maxsize.nendings[fl])
        return ERR_TOOBIG;      /* too many endings */
      for (i = 0; i < nf; i++)
      {
        int l = strlen(forms[i]) + 1;
        if (z + l > d->maxsize.endbuflen)
          return ERR_TOOBIG;    /* ending buffer too small */
        strcpy(d->endbuf + z, forms[i]);
        d->endtab[fl][ending][i] = d->endbuf + z;
        z += l;
      }

      /* There will be no error, we can safely update globals. */
      sz->endbuflen = z;
      sz->nendings[fl]++;
    }
  }
  sz->stembuflen += len;
  d->enttab[sz->nentries].type = type;
  d->enttab[sz->nentries].stem = stem;
  d->enttab[sz->nentries].ending = ending;
  sz->nentries++;
  return 0;
}

/* Compares approximately in reverse alphabetical order. */
static int entcmp(const void *ent1, const void *ent2)
{
  const struct entry *e1 = (const struct entry *)ent1;
  const struct entry *e2 = (const struct entry *)ent2;
  int r = strcmp(e1->stem, e2->stem);
  if (r != 0)
    return -r;
  r = e1->type - e2->type;
  if (r != 0)
    return r;
  return e1->ending - e2->ending;
}

static void dict_sort(dict *d, struct dictsize *sz)
{
  int i, j;

  if (sz->nentries <= 1)
    return;

  qsort(d->enttab, (size_t)sz->nentries, sizeof(d->enttab[0]), entcmp);

  /* Remove duplicated entries. */
  i = 0;
  for (j = 1; j < sz->nentries; j++)
    if (entcmp(&d->enttab[j], &d->enttab[i]))
      d->enttab[++i] = d->enttab[j];
  sz->nentries = i + 1;
}

static int dict_write(dict *d, const struct dictsize *sz, FILE *f)
{
  int i, j, k, a;

  if (fwrite(wfmark, sizeof(wfmark), 1, f) < 1)
    return ERR_WRITE;
  if (fwrite(sz, sizeof(*sz), 1, f) < 1)
    return ERR_WRITE;
  for (i = 0; i < sz->nentries; i++)
  {
    a = d->enttab[i].stem - d->stembuf;
    if (fwrite(&d->enttab[i].type, sizeof(d->enttab[i].type), 1, f) < 1)
      return ERR_WRITE;
    if (fwrite(&a, sizeof(a), 1, f) < 1)
      return ERR_WRITE;
    if (fwrite(&d->enttab[i].ending, sizeof(d->enttab[i].ending), 1, f)
        < 1)
      return ERR_WRITE;
  }
  for (i = 0; i < NINFLS; i++)
    for (j = 0; j < sz->nendings[i]; j++)
      for (k = 0; k < nforms[i]; k++)
      {
        a = d->endtab[i][j][k] - d->endbuf;
        if (fwrite(&a, sizeof(a), 1, f) < 1)
          return ERR_WRITE;
      }
  if (fwrite(d->stembuf, sizeof(*d->stembuf),
             (size_t)sz->stembuflen, f) < (size_t)sz->stembuflen)
    return ERR_WRITE;
  if (fwrite(d->endbuf, sizeof(*d->endbuf),
             (size_t)sz->endbuflen, f) < (size_t)sz->endbuflen)
    return ERR_WRITE;
  return 0;
}

#if 0

/*
 * "Osobna forme wolacza w jezyku polskim i szeregu innych maja jedynie
 * rzeczowniki rodzaju meskiego i zenskiego w liczbie pojedynczej" - J.
 * Tokarski, Fleksja polska, Warszawa 2001, str. 107.
 *
 * Returns non-zero if the above statement is true.
 */
static int vocative(const int type, char forms[MAXFORMS][MAXLEN + 1])
{
  const int inflexion = infl[type];


  if (inflexion == -1 || inflexion > 5)
    return 1;
  if (type < 4)
    return strcmp(forms[7], forms[13]) == 0;
  switch (inflexion)
  {
    case 0:
    case 2:
      return strcmp(forms[0], forms[6]) == 0 &&
        strcmp(forms[7], forms[13]) == 0;
    case 1:
      return strcmp(forms[0], forms[6]) == 0 &&
        strcmp(forms[7], forms[13]) == 0 &&
        strcmp(forms[14], forms[20]) == 0 &&
        strcmp(forms[21], forms[27]) == 0 &&
        strcmp(forms[28], forms[34]) == 0 &&
        strcmp(forms[35], forms[41]) == 0;
    case 3:
    case 4:
      return strcmp(forms[0], forms[6]) == 0 &&
        strcmp(forms[7], forms[13]) == 0 &&
        strcmp(forms[14], forms[20]) == 0 &&
        strcmp(forms[21], forms[27]) == 0;
    case 5:
      return strcmp(forms[0], forms[6]) == 0;
    default:
      fprintf(stderr, "%d %d\n", type, inflexion);
      assert(0);
  }
  return 1;
}

#endif

/* Compiles a dictionary. */
int dict_compile(FILE *f, FILE *g, int *line)
{
  dict *d;
  struct dictsize sz = {0, {0, 0, 0, 0, 0, 0, 0, 0}, 0, 0};
  int r, type;
  char forms[MAXFORMS][MAXLEN + 1];

  if ((d = dict_alloc(&maxsz)) == NULL)
    return ERR_NOMEM;

  *line = 0;
  for (;;)
  {
    if ((r = get_entry(f, &type, forms, line)) != 0)
      break;
    if (type < 0)
      break;
#if 0
    if (!vocative(type, forms))
      fprintf(stderr, "%s\n", forms[0]);
#endif
    if ((r = add_entry(type, forms, d, &sz)) != 0)
      break;
  }
  if (r == 0)
  {
    if (ferror(f))
      r = ERR_READ;
    else
    {
      dict_sort(d, &sz);
      r = sz.nentries > 0 ? dict_write(d, &sz, g) : ERR_BADSWF;
    }
  }
  dict_free(d);
  return r;
}

/* Reads contents of a dictionary from a word file. */
static int read_cont(FILE *f, dict *d)
{
  int i, j, k, a;

  for (i = 0; i < d->maxsize.nentries; i++)
  {
    if (fread(&d->enttab[i].type,
              sizeof(d->enttab[i].type), 1, f) < 1)
      return ERR_READ;
    if (fread(&a, sizeof(a), 1, f) < 1)
      return ERR_READ;
    if (a < 0 || a >= d->maxsize.stembuflen)
      return ERR_BADWF;         /* word file corrupted */
    d->enttab[i].stem = d->stembuf + a;
    if (fread(&d->enttab[i].ending,
              sizeof(d->enttab[i].ending), 1, f) < 1)
      return ERR_READ;
  }

  for (i = 0; i < NINFLS; i++)
    for (j = 0; j < d->maxsize.nendings[i]; j++)
      for (k = 0; k < nforms[i]; k++)
      {
        if (fread(&a, sizeof(a), 1, f) < 1)
          return ERR_READ;
        if (a < 0 || a >= d->maxsize.endbuflen)
          return ERR_BADWF;    /* word file corrupted */
        d->endtab[i][j][k] = d->endbuf + a;
      }

  if (fread(d->stembuf,
            sizeof(*d->stembuf),
            (size_t)d->maxsize.stembuflen,
            f) < (size_t)d->maxsize.stembuflen)
    return ERR_READ;

  if (fread(d->endbuf,
            sizeof(*d->endbuf),
            (size_t)d->maxsize.endbuflen,
            f) < (size_t)d->maxsize.endbuflen)
    return ERR_READ;

  return 0;
}

/* Reads dictionary from a word file. */
int dict_read(FILE *f, dict **d)
{
  struct dictsize sz;
  int i, r;
  char s[sizeof(wfmark)];

  *d = NULL;

  /* Check signature. */
  if (fread(&s, sizeof(s), 1, f) < 1)
    return ERR_READ;
  if (strcmp(s, wfmark))
    return ERR_BADWF;

  /* Read and check size of dictionary. */
  if (fread(&sz, sizeof(sz), 1, f) < 1)
    return ERR_READ;
  if (sz.nentries <= 0 || sz.nentries > maxsz.nentries)
    return ERR_BADWF;
  for (i = 0; i < NINFLS; i++)
    if (sz.nendings[i] < 0 || sz.nendings[i] > maxsz.nendings[i])
      return ERR_BADWF;
  if (sz.stembuflen < 0 || sz.stembuflen > maxsz.stembuflen)
    return ERR_BADWF;
  if (sz.endbuflen < 0 || sz.endbuflen > maxsz.endbuflen)
    return ERR_BADWF;

  /* Allocate memory and read contents. */
  if ((*d = dict_alloc(&sz)) == NULL)
    return ERR_NOMEM;
  (*d)->maxsize = sz;
  if ((r = read_cont(f, *d)) == 0)
    return 0;
  dict_free(*d);
  *d = NULL;
  return r;
}

/* Decompiles a dictionary to a file. */
int dict_decompile(FILE *f, FILE *g)
{
  dict *d;
  int r, type, fl, nf, i, j;

  if ((r = dict_read(f, &d)) != 0)
    return r;
  r = ERR_WRITE;
  for (i = 0; i < d->maxsize.nentries; i++)
  {
    type = d->enttab[i].type;
    fl = infl[type];
    nf = fl == -1 ? 1 : nforms[fl];
    if (fprintf(g, "%s\n", head[type]) < 0)
      goto fin;
    if (nf == 1)
    {
      if (fprintf(g, "%s\n", d->enttab[i].stem) < 0)
        goto fin;
    }
    else
      for (j = 0; j < nf; j++)
        if (strcmp(d->endtab[fl][d->enttab[i].ending][j], noform))
        {
          if (fprintf(g, "%s%s\n", d->enttab[i].stem,
                      d->endtab[fl][d->enttab[i].ending][j]) < 0)
            goto fin;
        }
        else
        {
          if (fprintf(g, "%s\n", noform) < 0)
            goto fin;
        }
    if (fprintf(g, "\n") < 0)
      goto fin;
  }
  r = 0;
fin:
  dict_free(d);
  return r;
}

static void get_desc(const dict *d, const struct entry *e,
                     int ending, dictdesc *dd)
{
  static const int cases[7] =
    {
      NOMINATIVE, GENITIVE, DATIVE, ACCUSATIVE, INSTRUMENTAL, LOCATIVE,
      VOCATIVE
    };
  static const int numbers[2] =
    {
      SINGULAR, PLURAL
    };

  const int type = e->type;
  const int fl = infl[type];
  const int nf = fl == -1 ? 1 : nforms[fl];
  lsym *sym = &dd->ls;
  const char *s;
  int i;

  /* main form first */
  if (nf == 1)                  /* uninflected entries */
  {
    if (strlen(e->stem) > MAXLEN)
      abort();                  /* dictionary corrupted */
    strcpy(dd->mf, e->stem);
  }
  else                          /* inflected entries */
  {
    s = NULL;                   /* ending of first non-empty form */
    for (i = 0; i < nf; i++)
      if (strcmp(d->endtab[fl][e->ending][i], noform))
      {
        s = d->endtab[fl][e->ending][i];
        if (strlen(e->stem) + strlen(s) > MAXLEN)
          abort();              /* dictionary corrupted */
        strcat(strcpy(dd->mf, e->stem), s);
        break;
      }
    if (s == NULL)
      abort();                  /* dictionary corrupted */
  }

  /* now symbol */
  *sym = lexical_info[type];
  /* set some fields of symbol according to position in the order of
   each inflexion */
  if (sym->i == SUBSTANTIVAL_DECLENSION)
  {
    assert(0 <= ending && ending < nforms[0]);
    sym->c = cases[ending % 7];
    sym->n = numbers[ending / 7];
  }
  else if (sym->i == CONJUGATION)
  {
    static const int persons[3] =
      {
        PERSON1, PERSON2, PERSON3
      };
    static const int genders[5] =
      {
        MASCULINE, FEMININE, NEUTER, MASCULINE_PERSONAL,
        NON_MASCULINE_PERSONAL
      };

    sym->i = 0;
    assert(0 <= ending && ending < nforms[7]);
    if (ending == 0)
      sym->fov = INFINITIVE;
    else if (ending < 7)
    {
      sym->fov = SIMPLE_PRESENT_OR_FUTURE;
      sym->i = GENDERLESS_CONJUGATION;
      ending--;
      sym->n = numbers[ending / 3];
      sym->p = persons[ending % 3];
    }
    else if (ending < 22)
    {
      sym->fov = PAST;
      sym->i = CONJUGATION;
      ending -= 7;
      sym->n = numbers[ending / 9];
      sym->g = genders[ending / 3];
      sym->p = persons[ending % 3];
    }
    else if (ending < 25)
    {
      sym->fov = IMPERATIVE_MOOD;
      sym->i = SIMPLE_FORMS_OF_IMPERATIVE;
      ending -= 22;
      sym->n = ending == 0 ? SINGULAR : PLURAL;
      sym->p = ending == 1 ? PERSON1 : PERSON2;
    }
    else if (ending < 40)
    {
      sym->fov = SUBJUNCTIVE_MOOD;
      sym->i = CONJUGATION;
      ending -= 25;
      sym->n = numbers[ending / 9];
      sym->g = genders[ending / 3];
      sym->p = persons[ending % 3];
    }
    else if (ending < 41)
      sym->fov = IMPERSONAL_FORM_OF_PAST_TENSE;
    else                        /* ending == 41 */
      sym->fov = ADVERBIAL_PARTICIPLE;
  }
  else if (sym->i == ADJECTIVAL_DECLENSION)
  {
    static const int genders[6] =
      {
        MASCULINE_ANIMATE, MASCULINE_INANIMATE, FEMININE, NEUTER,
        MASCULINE_PERSONAL, NON_MASCULINE_PERSONAL
      };

    assert(0 <= ending && ending < nforms[1]);
    sym->c = cases[ending % 7];
    sym->n = numbers[ending / 28];
    sym->g = genders[ending / 7];
  }
  else if (sym->i == NUMERAL_DECLENSION)
  {
    static const int genders[2] =
      {
        NON_MASCULINE_PERSONAL, MASCULINE_PERSONAL
      };

    assert(0 <= ending && ending < nforms[2]);
    sym->c = cases[ending % 7];
    sym->g = genders[ending / 7];
  }
  else if (sym->i == DECLENSION_OF_NUMERAL_1)
  {
    static const int genders[4] =
      {
        MASCULINE_ANIMATE, MASCULINE_INANIMATE, FEMININE, NEUTER
      };

    assert(0 <= ending && ending < nforms[3]);
    sym->c = cases[ending % 7];
    sym->g = genders[ending / 7];
  }
  else if (sym->i == DECLENSION_OF_NUMERAL_2)
  {
    static const int genders[4] =
      {
        MASCULINE_IMPERSONAL, MASCULINE_PERSONAL, FEMININE, NEUTER
      };

    assert(0 <= ending && ending < nforms[4]);
    sym->c = cases[ending % 7];
    sym->g = genders[ending / 7];
  }
  else if (sym->i == SINGULAR_DECLENSION_BY_CASE)
  {
    assert(0 <= ending && ending < nforms[5]);
    sym->c = cases[ending % 7];
  }
  else if (sym->i == SINGULAR_DECLENSION_BY_GENDER)
  {
    static const int genders[2] =
      {
        NON_FEMININE, FEMININE
      };
    assert(0 <= ending && ending < nforms[6]);
    sym->g = genders[ending];
  }

  /*
   * Here some corrections are made to sym in order to be in agreement
   * with lsym_isvalid. The table lexical_info contains information on
   * inflexion to be used in get_desc, but this is incorrect for
   * lsym_isvalid.
   */
  if (sym->s == NOUN ||
      sym->s == ADJECTIVE ||
      (sym->s == NUMERAL && (sym->ton == COLLECTIVE ||
                             sym->ton == ORDINAL)) ||
      sym->s == VERB ||
      sym->s == ADJECTIVAL_PASSIVE_PARTICIPLE ||
      sym->s == ADJECTIVAL_ACTIVE_PARTICIPLE ||
      sym->s == ADJECTIVAL_PAST_PARTICIPLE ||
      sym->s == VERBAL_NOUN)
    sym->i = 0;
  if (sym->s == NUMERAL && sym->ton == CARDINAL &&
      sym->i == SUBSTANTIVAL_DECLENSION)
    sym->g = strcmp(dd->mf, "zero") ? MASCULINE_INANIMATE : NEUTER;
}

/*
 * Adds description to the list if not present. Returns: 0 - no memory,
 * 1 - a description already exists, 2 - a new description added
 * successfully.
 */
static int add_desc(const dict *d, const struct entry *e,
                    int ending, dictdesc **dd)
{
  dictdesc tdd, *t, *tnew;
  get_desc(d, e, ending, &tdd);
  for (t = *dd; t != NULL; t = t->next)
    if (strcmp(tdd.mf, t->mf) == 0 && lsym_eq(&tdd.ls, &t->ls))
      return 1;
  if ((tnew = NEW(dictdesc)) == NULL)
    return 0;
  *tnew = tdd;
  *dd = sl_push(*dd, tnew);
  return 2;
}

/*
 * Finds descriptions of a word in a dictionary. Expands the list of
 * descriptions with new descriptions, if they exist. Returns 0 if ok
 * (also if nothing has been added to the list). Returns ERR_NOMEM only
 * if there has been something to add which was not added.
 */
int dict_find(const dict *d, const char *w, dictdesc **dd)
{
  const int ne = d->maxsize.nentries;
  const size_t len = strlen(w);
  char buf[MAXLEN + 1 + 1];     /* copy of w plus walking sentinel */
  char *s = strcpy(buf, w);     /* stem */
  char *b = s + len;            /* walking sentinel at end of stem */
  char *e = b + 1;              /* ending */
  int l, u, m;
  int i, j, fl, nf, en;

  if (ne <= 0)
    return 0;
  if (len == 0 || len >= MAXLEN)
    return 0;

  /* Search dictionary dividing word into stem s and ending e. */
  *e = '\0';
  while (b >= s)
  {
    /* Binary searching in a non-decreasingly ordered table of strings. */
    l = -1;
    u = ne;
    while (l + 1 != u)
    {
      m = (l + u) / 2;
      if (strcmp(d->enttab[m].stem, s) > 0)
        l = m;
      else
        u = m;
    }

    /* Linear search forward. */
    for (i = u; i < ne && strcmp(d->enttab[i].stem, s) == 0; i++)
    {
      fl = infl[(int)d->enttab[i].type];
      nf = fl == -1 ? 1 : nforms[fl];
      if (nf == 1)              /* uninflected entries */
      {
        if (*e == '\0')
          if (!add_desc(d, &d->enttab[i], -1, dd))
            return ERR_NOMEM;
      }
      else                      /* inflected entries */
      {
        en = d->enttab[i].ending;
        for (j = 0; j < nf; j++)
          if (strcmp(d->endtab[fl][en][j], e) == 0)
            if (!add_desc(d, &d->enttab[i], j, dd))
              return ERR_NOMEM;
      }
    }
    *--e = *--b;
    *b = '\0';
  }
  return 0;
}

/**************************************************************
 *                                                            *
 *                    Get lexical symbol.                     *
 *                                                            *
 **************************************************************/

/* Some ISO_8859-2:1987 codes. */
#define CHAR_LF                    10
#define CHAR_EXCLAMATION_MARK      33
#define CHAR_QUOTATION_MARK        34
#define CHAR_NUMBER_SIGN           35
#define CHAR_DOLLAR_SIGN           36
#define CHAR_PERCENT_SIGN          37
#define CHAR_AMPERSAND             38
#define CHAR_APOSTROPHE            39
#define CHAR_LEFT_PARENTHESIS      40
#define CHAR_RIGHT_PARENTHESIS     41
#define CHAR_ASTERISK              42
#define CHAR_PLUS_SIGN             43
#define CHAR_COMMA                 44
#define CHAR_HYPHEN_MINUS          45
#define CHAR_FULL_STOP             46
#define CHAR_SOLIDUS               47
#define CHAR_COLON                 58
#define CHAR_SEMICOLON             59
#define CHAR_LESS_THAN_SIGN        60
#define CHAR_EQUALS_SIGN           61
#define CHAR_GREATER_THAN_SIGN     62
#define CHAR_QUESTION_MARK         63
#define CHAR_COMMERCIAL_AT         64
#define CHAR_LEFT_SQUARE_BRACKET   91
#define CHAR_REVERSE_SOLIDUS       92
#define CHAR_RIGHT_SQUARE_BRACKET  93
#define CHAR_CIRCUMFLEX            94
#define CHAR_LOW_LINE              95
#define CHAR_GRAVE                 96
#define CHAR_LEFT_CURLY_BRACKET   123
#define CHAR_VERTICAL_LINE        124
#define CHAR_RIGHT_CURLY_BRACKET  125
#define CHAR_TILDE                126
#define CHAR_SECTION_SIGN         ((unsigned char)(167))

#define SET_PM(x)             \
{                             \
  ls->s = PUNCTUATION_MARK;   \
  ls->pm = x;                 \
  s[i++] = c;                 \
}

#define SET_GS(x)             \
{                             \
  ls->s = GRAPHIC_SIGN;       \
  ls->gs = x;                 \
  s[i++] = c;                 \
}

static void getsym1(FILE *f, char *s, int maxlen, lsym *ls,
                    int *line, int *is_hanging)
{
  int i = 0, c;

  assert(maxlen >= 3);          /* must hold three dots */

  lsym_clear(ls);

  if (*is_hanging)
  {
    ls->s = PUNCTUATION_MARK;
    ls->pm = FULL_STOP;
    s[i++] = CHAR_FULL_STOP;
    s[i] = '\0';
    *is_hanging = 0;
    return;
  }

  while (l2_isspace(c = fgetc(f)))
    if (c == CHAR_LF)
      (*line)++;

  if (l2_isalpha(c))
  {
    ls->s = UNKNOWN_WORD;
    do
      if (i < maxlen)
        s[i++] = c;
    while (l2_isalpha(c = fgetc(f)));
    ungetc(c, f);
  }
  else if (c == CHAR_FULL_STOP)
  {
    ls->s = PUNCTUATION_MARK;
    ls->pm = FULL_STOP;
    s[i++] = CHAR_FULL_STOP;
    if ((c = fgetc(f)) != CHAR_FULL_STOP)
      ungetc(c, f);
    else if ((c = fgetc(f)) != CHAR_FULL_STOP)
    {
      ungetc(c, f);
      *is_hanging = 1;
    }
    else
    {
      ls->pm = DOTS;
      s[i++] = CHAR_FULL_STOP;
      s[i++] = CHAR_FULL_STOP;
    }
  }
  else if (c == CHAR_SEMICOLON)
    SET_PM(SEMICOLON)
      else if (c == CHAR_COMMA)
      {
        ls->s = PUNCTUATION_MARK;
        s[i++] = CHAR_COMMA;
        if ((c = fgetc(f)) == CHAR_COMMA)
        {
          ls->pm = LEFT_QUOTATION_MARK;
          s[i++] = CHAR_COMMA;
        }
        else
        {
          ls->pm = COMMA;
          ungetc(c, f);
        }
      }
  else if (c == CHAR_COLON)
    SET_PM(COLON)
      else if (c == CHAR_HYPHEN_MINUS)
      {
        s[i++] = CHAR_HYPHEN_MINUS;
        if ((c = fgetc(f)) == CHAR_HYPHEN_MINUS)
        {
          ls->s = PUNCTUATION_MARK;
          ls->pm = DASH;
          s[i++] = CHAR_HYPHEN_MINUS;
        }
        else
        {
          ungetc(c, f);
          ls->s = HYPHEN;
        }
      }
  else if (c == CHAR_QUESTION_MARK)
    SET_PM(QUESTION_MARK)
  else if (c == CHAR_EXCLAMATION_MARK)
    SET_PM(EXCLAMATION_MARK)
  else if (c == CHAR_LEFT_PARENTHESIS)
    SET_PM(LEFT_ROUND_BRACKET)
  else if (c == CHAR_RIGHT_PARENTHESIS)
    SET_PM(RIGHT_ROUND_BRACKET)
  else if (c == CHAR_LEFT_SQUARE_BRACKET)
    SET_PM(LEFT_SQUARE_BRACKET)
  else if (c == CHAR_RIGHT_SQUARE_BRACKET)
    SET_PM(RIGHT_SQUARE_BRACKET)
  else if (c == CHAR_SOLIDUS)
    SET_PM(SIMPLE_BRACKET)
  else if (c == CHAR_LEFT_CURLY_BRACKET)
    SET_PM(LEFT_CURLY_BRACKET)
  else if (c == CHAR_RIGHT_CURLY_BRACKET)
    SET_PM(RIGHT_CURLY_BRACKET)
  else if (c == CHAR_LESS_THAN_SIGN)
  {
    ls->s = PUNCTUATION_MARK;
    s[i++] = CHAR_LESS_THAN_SIGN;
    if ((c = fgetc(f)) == CHAR_LESS_THAN_SIGN)
    {
      ls->pm = LEFT_ANGLE_QUOTATION_MARK;
      s[i++] = CHAR_LESS_THAN_SIGN;
    }
    else
    {
      ls->pm = LEFT_ANGLE_BRACKET;
      ungetc(c, f);
    }
  }
  else if (c == CHAR_GREATER_THAN_SIGN)
  {
    ls->s = PUNCTUATION_MARK;
    s[i++] = CHAR_GREATER_THAN_SIGN;
    if ((c = fgetc(f)) == CHAR_GREATER_THAN_SIGN)
    {
      ls->pm = RIGHT_ANGLE_QUOTATION_MARK;
      s[i++] = CHAR_GREATER_THAN_SIGN;
    }
    else
    {
      ls->pm = RIGHT_ANGLE_BRACKET;
      ungetc(c, f);
    }
  }
  else if (c == CHAR_APOSTROPHE)
  {
    ls->s = PUNCTUATION_MARK;
    s[i++] = CHAR_APOSTROPHE;
    if ((c = fgetc(f)) == CHAR_APOSTROPHE)
    {
      ls->pm = RIGHT_QUOTATION_MARK;
      s[i++] = CHAR_APOSTROPHE;
    }
    else
    {
      ls->pm = RIGHT_DEFINITION_QUOTATION_MARK;
      ungetc(c, f);
    }
  }
  else if (c == CHAR_GRAVE)
    SET_PM(LEFT_DEFINITION_QUOTATION_MARK)
  else if (c == CHAR_QUOTATION_MARK)
    SET_PM(TYPEWRITER_QUOTATION_MARK)
  else if (l2_isdigit(c))
  {
    ls->s = NUMBER;
    do
      if (i < maxlen)
        s[i++] = c;
    while (l2_isdigit(c = fgetc(f)));
    ungetc(c, f);
  }
  else if (c == CHAR_EQUALS_SIGN)
    SET_GS(EQUALS_SIGN)
  else if (c == CHAR_PLUS_SIGN)
    SET_GS(PLUS_SIGN)
  else if (c == CHAR_PERCENT_SIGN)
    SET_GS(PERCENT_SIGN)
  else if (c == CHAR_SECTION_SIGN)
    SET_GS(PARAGRAPH_SIGN)
  else if (c == CHAR_NUMBER_SIGN)
    SET_GS(NUMBER_SIGN)
  else if (c == CHAR_DOLLAR_SIGN)
    SET_GS(DOLLAR_SIGN)
  else if (c == CHAR_AMPERSAND)
    SET_GS(AMPERSAND)
  else if (c == CHAR_ASTERISK)
    SET_GS(ASTERISK)
  else if (c == CHAR_COMMERCIAL_AT)
    SET_GS(COMMERCIAL_AT)
  else if (c == CHAR_REVERSE_SOLIDUS)
    SET_GS(BACKSLASH)
  else if (c == CHAR_CIRCUMFLEX)
    SET_GS(UP_ARROW)
  else if (c == CHAR_LOW_LINE)
    SET_GS(UNDERLINE)
  else if (c == CHAR_VERTICAL_LINE)
    SET_GS(VERTICAL_LINE)
  else if (c == CHAR_TILDE)
    SET_GS(TILDE)
  else if (c != EOF)
    ls->s = UNKNOWN_CHARACTER;
  s[i] = '\0';
}

#undef SET_PM
#undef SET_GS

static int search(const dictlist *dl, const char *w, dictdesc **dd)
{
  for (; dl != NULL; dl = dl->next)
    switch (dict_find(dl->d, w, dd))
    {
      case 0:
        break;
      case ERR_NOMEM:
        return 1;
      default:
        assert(0);
    }
  return 0;
}

/* True if description d exists in the list dl. */
static int dpresent(const dictdesc *dl, const dictdesc *d)
{
  const dictdesc *t;
  for (t = dl; t != NULL; t = t->next)
    if (strcmp(t->mf, d->mf) == 0 && lsym_eq(&t->ls, &d->ls))
      return 1;
  return 0;
}

/* w == "nie" + a word? Returns: 0 = success, 1 = ERR_NOMEM. */
static int chknie(const dictlist *dl, const char *w, dictdesc **dd)
{
  dictdesc *dd1 = NULL, *tdd;
  char *t;
  size_t len;
  int type;

  if (strncmp(w, "nie", 3))
    return 0;
  w += 3;
  if (!*w)
    return 0;
  if (search(dl, w, &dd1))
    goto fin;

  while ((tdd = sl_pop(&dd1)) != NULL)
    if ((type = tdd->ls.s) == NOUN ||
        (type == ADJECTIVE && tdd->ls.d == POSITIVE) ||
        type == ADJECTIVAL_PASSIVE_PARTICIPLE ||
        type == ADJECTIVAL_ACTIVE_PARTICIPLE ||
        type == VERBAL_NOUN ||
        type == ADJECTIVAL_PAST_PARTICIPLE ||
        (type == ADVERB && tdd->ls.d == POSITIVE))
    {
      t = tdd->mf;
      len = strlen(t);
      assert(len + 3 <= MAXLEN);
      memmove(t + 3, t, len + 1);
      t[0] = 'n'; t[1] = 'i'; t[2] = 'e';
      if (!dpresent(*dd, tdd))
        *dd = sl_push(*dd, tdd);
      else
        free(tdd);
    }
    else
      free(tdd);

  assert(dd1 == NULL);
  return 0;
fin:
  sl_free(dd1, NULL);
  return 1;
}

/*
 * w == "naj" + adjective or adverb? Returns: 0 = success, 1 =
 * ERR_NOMEM.
 */
static int chknaj(const dictlist *dl, const char *w, dictdesc **dd)
{
  dictdesc *dd1 = NULL, *tdd;
  char *t;
  size_t len;
  int type;

  if (strncmp(w, "naj", 3))
    return 0;
  w += 3;
  if (!*w)
    return 0;
  if (search(dl, w, &dd1))
    goto fin;

  while ((tdd = sl_pop(&dd1)) != NULL)
    if (((type = tdd->ls.s) == ADJECTIVE || type == ADVERB) &&
        tdd->ls.d == COMPARATIVE)
    {
      t = tdd->mf;
      len = strlen(t);
      assert(len + 3 <= MAXLEN);
      memmove(t + 3, t, len + 1);
      t[0] = 'n'; t[1] = 'a'; t[2] = 'j';
      tdd->ls.d = SUPERLATIVE;
      if (!dpresent(*dd, tdd))
        *dd = sl_push(*dd, tdd);
      else
        free(tdd);
    }
    else
      free(tdd);
  assert(dd1 == NULL);
  return 0;
fin:
  sl_free(dd1, NULL);
  return 1;
}

/*
 * Checks if w is a word ended with the particle -ze, -z. Examples:
 * idez, idzze, onze, myz, oniz, zyjez on?, konze to?
 */
static int chkze(const dictlist *dl, char *w, dictdesc **dd)
{
  size_t l = strlen(w);
  dictdesc *dd1 = NULL, *tdd;
  char c;
  int r;

  if (l > 3 && w[l - 1] == 'e' && w[l - 2] == '\277' &&
      l2_isconsonant(w[l - 3]))
    l -= 2;
  else if (l > 2 && w[l - 1] == '\277' && l2_isvowel(w[l - 2]))
    l -= 1;
  else
    return 0;

  c = w[l]; w[l] = '\0';
  r = search(dl, w, &dd1);
  w[l] = c;
  if (r)
    goto fin;
  c = w[l]; w[l] = '\0';
  r = chknie(dl, w, &dd1);
  w[l] = c;
  if (r)
    goto fin;
  c = w[l]; w[l] = '\0';
  r = chknaj(dl, w, &dd1);
  w[l] = c;
  if (r)
    goto fin;

  while ((tdd = sl_pop(&dd1)) != NULL)
  {
    tdd->ls.pz = ZE_PRESENT;
    if (!dpresent(*dd, tdd))
      *dd = sl_push(*dd, tdd);
    else
      free(tdd);
  }
  assert(dd1 == NULL);
  return 0;
fin:
  sl_free(dd1, NULL);
  return 1;
}

/*
 * Checks if w is a non-verb ended with the personal ending -(e)m,
 * -(e)s, -(e)smy, -(e)scie and eventually the particle -ze, -z.
 * Examples: mlekom/mlekozem/mlekomze pil, rumem/rumzem/rumemze pil,
 * nieladniesmy/nieladniezesmy/nieladniesmyz postapili. Each possible
 * form is a concatenation of a prefix and an ending tab[i].suffix.
 */
static int chkem(const dictlist *dl, char *w, dictdesc **dd)
{
  static const struct {char *suffix; size_t len;}
  tab[] =
    {
      /* prefix ends with a vowel */
      {"m", 1},     {"\266", 1},
      {"\266my", 3},   {"\266cie", 4},
      {"m\277e", 3},   {"\266\277e", 3},
      {"\266my\277", 4},  {"\266cie\277", 5},
      /* prefix ends with a consonant */
      {"em", 2},    {"e\266", 2},
      {"e\266my", 4},  {"e\266cie", 5},
      {"em\277e", 4},  {"e\266\277e", 4},
      {"e\266my\277", 5}, {"e\266cie\277", 6},
      /* every ending of prefix */
      {"\277em", 3},   {"\277e\266", 3},
      {"\277e\266my", 5}, {"\277e\266cie", 6},
      {NULL, 0}
    };
  size_t l0 = strlen(w), l;
  int i, j, r;
  dictdesc *dd1 = NULL, *tdd;
  char c;

  for (i = 0; tab[i].suffix != NULL; i++)
    if (ispsuffix2(w, l0, tab[i].suffix, tab[i].len))
    {
      assert(tab[i].len < l0);
      l = l0 - tab[i].len;
      if (i >= 16 ||
          (i >= 8 && l2_isconsonant(w[l - 1])) ||
          l2_isvowel(w[l - 1]))
      {
        c = w[l]; w[l] = '\0';
        r = search(dl, w, &dd1);
        w[l] = c;
        if (r)
          goto fin;
        c = w[l]; w[l] = '\0';
        r = chknie(dl, w, &dd1);
        w[l] = c;
        if (r)
          goto fin;
        c = w[l]; w[l] = '\0';
        r = chknaj(dl, w, &dd1);
        w[l] = c;
        if (r)
          goto fin;

        /* Remove new items if they are of unproper part of speech. Do
         not adjust main forms. */
        while ((tdd = sl_pop(&dd1)) != NULL)
          if (tdd->ls.s != VERB)
          {
            j = i / 4;
            if (j == 1 || j > 2)
              tdd->ls.pz = ZE_PRESENT;
            tdd->ls.pe = petab[i % 4 + 1];
            if (!dpresent(*dd, tdd))
              *dd = sl_push(*dd, tdd);
            else
              free(tdd);
          }
          else
            free(tdd);
        assert(dd1 == NULL);

      }
    }
  return 0;

fin:
  sl_free(dd1, NULL);
  return 1;
}

/* Requires char *w, but does not change *w. */
static int hsearch(const dictlist *dl, char *w, dictdesc **dd)
{
  if (search(dl, w, dd))
    return 1;
  if (chknie(dl, w, dd))
    return 1;
  if (chknaj(dl, w, dd))
    return 1;
  if (chkze(dl, w, dd))
    return 1;
  if (chkem(dl, w, dd))
    return 1;
  return 0;
}

/*
 * Gets next lexical symbol from the file f. The list dl should be a
 * list of dictionaries to use. The variable is_hanging contains
 * internal state of the function getsym and should be set to zero
 * before the first call on the newly rewound file. The result is
 * returned as a list. If the result is NULL, either end of file has
 * been reached, an I/O error occured or there is not enough memory:
 *
 * if (getsym == NULL)
 * {
 *   if (ferror(f))
 *     io_error();
 *   else if (feof(f))
 *     end_of_file();
 *   else
 *     not_enough_memory()
 * }
 */
dictdesc *getsym(FILE *f, char *s, int *line, int *is_hanging,
                 const dictlist *dl)
{
  lsym ls;
  dictdesc *dd = NULL, *tdd;
  char w[MAXLEN + 1];

  getsym1(f, s, MAXLEN, &ls, line, is_hanging);

  if (ls.s == UNKNOWN_WORD)
  {
    assert(*s);
    strcpy(w, s);

    if (hsearch(dl, w, &dd))
      goto fin;

    /*
     * Because one-letter words in upper case are both capitalized and
     * in upper case, the order of l2_iscapit_str and l2_isupper_str
     * below is important.
     */

    if (l2_islower_str(w))
      ;
    else if (l2_iscapit_str(w))
    {
      if (hsearch(dl, l2_tolower_str(w), &dd))
        goto fin;
    }
    else if (l2_isupper_str(w))
    {
      if (hsearch(dl, l2_tocapit_str(w), &dd))
        goto fin;
      if (hsearch(dl, l2_tolower_str(w), &dd))
        goto fin;
    }
    else                        /* mixed case */
    {
      if (hsearch(dl, l2_toupper_str(w), &dd))
        goto fin;
      if (hsearch(dl, l2_tocapit_str(w), &dd))
        goto fin;
      if (hsearch(dl, l2_tolower_str(w), &dd))
        goto fin;
    }

    /*
     * Here we have w in lower case, but it is not used any more.
     */

    if (dd == NULL)             /* nothing is found: UNKNOWN_WORD */
    {
      if ((tdd = NEW(dictdesc)) == NULL)
        goto fin;
      strcpy(tdd->mf, s);
      tdd->ls = ls;
      dd = sl_push(dd, tdd);
    }
  }
  else if (ls.s == PUNCTUATION_MARK ||
           ls.s == HYPHEN ||
           ls.s == NUMBER ||
           ls.s == GRAPHIC_SIGN ||
           ls.s == UNKNOWN_CHARACTER)
  {
    assert((ls.s == UNKNOWN_CHARACTER && !*s)|| *s);
    if ((tdd = NEW(dictdesc)) == NULL)
      goto fin;
    strcpy(tdd->mf, s);
    tdd->ls = ls;
    dd = sl_push(dd, tdd);
  }
  else
    assert(ls.s == 0 && (!*s));

  return dd;
fin:
  sl_free(dd, NULL);
  return NULL;
}

/**************************************************************
 *                                                            *
 *                      Driver routines.                      *
 *                                                            *
 **************************************************************/

/* Frees a node of a list of dictionaries. To be used with sl_free. */
static void free_dl(void *dl)
{
  dict_free(((dictlist *)dl)->d);
  free(dl);
}

int pla_run(int argc, char *argv[])
{
  static const char *const enomem = "not enough memory";
  const char *progname = NULL;
  char s[1001];
  dictlist *dl = NULL;
  FILE *f, *g;
  int i, j;

  progname = argv[0];
  if (! ((argc >= 2) &&
         ((!strcmp(argv[1], "analyse") && argc >= 4) ||
          (!strcmp(argv[1], "compile") && argc == 4) ||
          (!strcmp(argv[1], "decompile") && argc == 4) ||
          (!strcmp(argv[1], "list-symbols") && argc == 2))))
  {
    printf("pla, version %s\nStanislaw Galus 2008\n",
           version);
    printf("\nUsage:\n");
    printf("pla analyse [word-file]... input-file output-file\n");
    printf("pla compile source-word-file word-file\n");
    printf("pla decompile word-file source-word-file\n");
    printf("pla list-symbols\n\n");
    return 0;
  }
  *s = '\0';
  for (i = 2; i < argc; i++)
    for (j = i + 1; j < argc; j++)
      if (!strcmp(argv[i], argv[j]))
      {
        sprintf(s, "duplicated file name %s", argv[i]);
        goto fin;
      }
  if (!strcmp(argv[1], "analyse"))
  {
    dict *d;
    dictlist *e;
    int line = 0, is_hanging = 0;
    dictdesc *dd, *t;
    char s[MAXLEN + 1];
    char lsb[MAXSYMLEN + 1];

    for (i = 2; i < argc - 2; i++)
    {
      if ((f = fopen(argv[i], "rb")) == NULL)
      {
        sprintf(s, "cannot open word file %s", argv[i]);
        goto fin;
      }
      if (dict_read(f, &d) != 0)
      {
        fclose(f);
        sprintf(s, "cannot read word file %s", argv[i]);
        goto fin;
      }
      fclose(f);
      if ((e = NEW(dictlist)) == NULL)
      {
        sprintf(s, "%s", enomem);
        goto fin;
      }
      e->d = d;
      dl = sl_push(dl, e);
    }
    if ((f = fopen(argv[argc - 2], "r")) == NULL)
    {
      sprintf(s, "cannot open input file %s", argv[argc - 2]);
      goto fin;
    }
    if ((g = fopen(argv[argc - 1], "w")) == NULL)
    {
      sprintf(s, "cannot open output file %s", argv[argc - 1]);
      fclose(f);
      goto fin;
    }
    while ((dd = getsym(f, s, &line, &is_hanging, dl)) != NULL)
    {
      assert(lsym_isvalid(&dd->ls));
      switch (dd->ls.s)
      {
        case UNKNOWN_CHARACTER:
          fprintf(g, "\n     %s\n", lsym_tostr(&dd->ls, lsb));
          break;
        case WORD_OF_NONE_PART_OF_SPEECH:
        case UNKNOWN_WORD:
        case PUNCTUATION_MARK:
        case HYPHEN:
        case NUMBER:
        case GRAPHIC_SIGN:
          fprintf(g, "%s\n     %s\n", s, lsym_tostr(&dd->ls, lsb));
          break;
        default:
          fprintf(g, "%s\n", s);
          for (t = dd; t != NULL; t = t->next)
            fprintf(g, "     %s %s\n", lsym_tostr(&t->ls, lsb), t->mf);
          break;
      }
      sl_free(dd, NULL);
    }
    if (ferror(f))
      sprintf(s, "%s", "cannot read input file");
    else if (!feof(f))
      sprintf(s, "%s", enomem);
    else if (ferror(g))
      sprintf(s, "%s", "cannot write output file");
    fclose(g);
    fclose(f);
  }
  else if (!strcmp(argv[1], "compile"))
  {
    int line;

    if ((f = fopen(argv[argc - 2], "r")) == NULL)
    {
      sprintf(s, "cannot open source word file %s", argv[argc - 2]);
      goto fin;
    }
    if ((g = fopen(argv[argc - 1], "wb")) == NULL)
    {
      sprintf(s, "cannot open word file %s", argv[argc - 1]);
      fclose(f);
      goto fin;
    }
    switch (dict_compile(f, g, &line))
    {
      case 0:
        break;
      case ERR_NOMEM:
        sprintf(s, "%s", enomem);
        break;
      case ERR_READ:
        sprintf(s, "%s", "cannot read source word file");
        break;
      case ERR_WRITE:
        sprintf(s, "%s", "cannot write word file");
        break;
      case ERR_TOOBIG:
        sprintf(s, "%s", "dictionary too big");
        break;
      case ERR_BADSWF:
        sprintf(s, "bad source word file, line %d", line);
        break;
      default:
        assert(0);
    }
    fclose(g);
    fclose(f);
  }
  else if (!strcmp(argv[1], "decompile"))
  {
    if ((f = fopen(argv[argc - 2], "rb")) == NULL)
    {
      sprintf(s, "cannot open word file %s", argv[argc - 2]);
      goto fin;
    }
    if ((g = fopen(argv[argc - 1], "w")) == NULL)
    {
      sprintf(s, "cannot open source word file %s", argv[argc - 1]);
      fclose(f);
      goto fin;
    }
    switch (dict_decompile(f, g))
    {
      case 0:
        break;
      case ERR_NOMEM:
        sprintf(s, "%s", enomem);
        break;
      case ERR_READ:
        sprintf(s, "%s", "cannot read word file");
        break;
      case ERR_WRITE:
        sprintf(s, "%s", "cannot write source word file");
        break;
      case ERR_BADWF:
        sprintf(s, "%s", "bad word file");
        break;
      default:
        assert(0);
    }
    fclose(g);
    fclose(f);
  }
  else
    lsym_genlist(stdout);

fin:
  sl_free(dl, free_dl);

  if (*s)
  {
    fflush(stdout);
    fprintf(stderr, "error: ");
    if (progname != NULL)
      fprintf(stderr, "%s: ", progname);
    fprintf(stderr, "%s\n", s);
    return 1;
  }
  return 0;
}

int main(int argc, char *argv[])
{
#ifdef MCHECK
  mtrace();
#endif
  return pla_run(argc, argv);
}
