/*
 * Interface to Polish Lexical Analyser
 * Stanislaw Galus
 * 1 November 2008
 */

/* $Id: pla.h,v 1.8 2008/11/01 06:14:28 sgalus Exp $ */

#ifndef PLA_H
#define PLA_H

#define ERR_NOMEM  1            /* not enough memory */
#define ERR_READ   2            /* error reading file */
#define ERR_WRITE  3            /* error writing file */
#define ERR_TOOBIG 4            /* dictionary too big */
#define ERR_BADSWF 5            /* bad source word file */
#define ERR_BADWF  6            /* bad word file */

#define MAXLEN 100              /* maximum length of a word */
#define MAXSYMLEN 40            /* see lsym_tostr */

extern const char *const version;

int l2_isalnum(int c);
int l2_isalpha(int c);
int l2_iscntrl(int c);
int l2_isdigit(int c);
int l2_isgraph(int c);
int l2_islower(int c);
int l2_isprint(int c);
int l2_ispunct(int c);
int l2_isspace(int c);
int l2_isupper(int c);
int l2_isxdigit(int c);
int l2_tolower(int c);
int l2_toupper(int c);

int l2_islower_str(const char *s);
int l2_isupper_str(const char *s);
int l2_iscapit_str(const char *s);
char *l2_tolower_str(char *s);
char *l2_toupper_str(char *s);
char *l2_tocapit_str(char *s);
int l2_isvowel(char c);
int l2_isconsonant(char c);

char *strdup(const char *s);
char *trim(char *s);
int iscat(const char *s, const char *t, const char *u);
int isprefix(const char *s1, const char *s2);
int issuffix(const char *s1, const char *s2);
int ispsuffix(const char *s1, const char *s2);
int ispsuffix2(const char *s1, size_t l1, const char *s2, size_t l2);
int chksum(const void *ptr, size_t size);
int chksumf(const char *fn);

typedef struct lsym             /* lexical symbol */
{
  int s;                        /* symbol */
  int top;                      /* type of pronoun */
  int ton;                      /* type of numeral */
  int fov;                      /* form of verb */
  int i;                        /* inflexion */
  int c;                        /* case */
  int n;                        /* number */
  int g;                        /* gender */
  int p;                        /* person */
  int d;                        /* degree */
  int a;                        /* aspect */
  int pm;                       /* punctuation mark */
  int gs;                       /* graphic sign */
  int pe;                       /* personal ending */
  int pz;                       /* particle ze */
} lsym;

enum                            /* symbols */
{
  NOUN = 1, ADJECTIVE, PRONOUN, NUMERAL, VERB,
  ADJECTIVAL_PASSIVE_PARTICIPLE, ADJECTIVAL_ACTIVE_PARTICIPLE,
  VERBAL_NOUN, ADJECTIVAL_PAST_PARTICIPLE, ADVERB, PREPOSITION,
  CONJUNCTION, PARTICLE, INTERJECTION, WORD_OF_NONE_PART_OF_SPEECH,
  UNKNOWN_WORD, PUNCTUATION_MARK, HYPHEN, NUMBER, GRAPHIC_SIGN,
  UNKNOWN_CHARACTER
};

enum                            /* types of pronouns */
{
  PERSONAL = 1, REFLEXIVE, INDEFINITE, NEGATIVE, DEMONSTRATIVE,
  POSSESSIVE, INTERROGATIVE_RELATIVE
};

enum                            /* types of numerals */
{
  CARDINAL = 1, COLLECTIVE, FRACTIONAL, ORDINAL
};

enum                            /* forms of verbs */
{
  INFINITIVE = 1, SIMPLE_PRESENT_OR_FUTURE, PAST, IMPERATIVE_MOOD,
  SUBJUNCTIVE_MOOD, IMPERSONAL_FORM_OF_PAST_TENSE,
  ADVERBIAL_PARTICIPLE
};

enum                            /* inflexions */
{
  SUBSTANTIVAL_DECLENSION = 1, ADJECTIVAL_DECLENSION,
  NUMERAL_DECLENSION, DECLENSION_OF_NUMERAL_1,
  DECLENSION_OF_NUMERAL_2, SINGULAR_DECLENSION_BY_CASE,
  SINGULAR_DECLENSION_BY_GENDER, CONJUGATION, GENDERLESS_CONJUGATION,
  SIMPLE_FORMS_OF_IMPERATIVE
};

enum                            /* declension cases */
{
  NOMINATIVE = 1, GENITIVE, DATIVE, ACCUSATIVE, INSTRUMENTAL, LOCATIVE,
  VOCATIVE
};

enum                            /* numbers */
{
  SINGULAR = 1, PLURAL
};

enum                            /* genders */
{
  MASCULINE = 1, MASCULINE_PERSONAL, MASCULINE_ANIMAL,
  MASCULINE_INANIMATE, FEMININE, NEUTER, MASCULINE_ANIMATE,
  MASCULINE_IMPERSONAL, NON_MASCULINE_PERSONAL, NON_FEMININE
};

enum                            /* persons */
{
  PERSON1 = 1, PERSON2, PERSON3
};

enum                            /* degrees */
{
  POSITIVE = 1, COMPARATIVE, SUPERLATIVE
};

enum                            /* aspects */
{
  IMPERFECT = 1, PERFECT
};

enum                            /* punctuation marks */
{
  FULL_STOP = 1, SEMICOLON, COMMA, COLON, DASH, DOTS, QUESTION_MARK,
  EXCLAMATION_MARK, LEFT_ROUND_BRACKET, RIGHT_ROUND_BRACKET,
  LEFT_SQUARE_BRACKET, RIGHT_SQUARE_BRACKET, SIMPLE_BRACKET,
  LEFT_CURLY_BRACKET, RIGHT_CURLY_BRACKET, LEFT_ANGLE_BRACKET,
  RIGHT_ANGLE_BRACKET, LEFT_QUOTATION_MARK, RIGHT_QUOTATION_MARK,
  LEFT_ANGLE_QUOTATION_MARK, RIGHT_ANGLE_QUOTATION_MARK,
  LEFT_DEFINITION_QUOTATION_MARK, RIGHT_DEFINITION_QUOTATION_MARK,
  TYPEWRITER_QUOTATION_MARK
};

enum                            /* graphic signs */
{
  EQUALS_SIGN = 1, PLUS_SIGN, PERCENT_SIGN, PARAGRAPH_SIGN, NUMBER_SIGN,
  DOLLAR_SIGN, AMPERSAND, ASTERISK, COMMERCIAL_AT, BACKSLASH, UP_ARROW,
  UNDERLINE, VERTICAL_LINE, TILDE
};

enum                            /* personal endings */
{
  PERSON1_SINGULAR = 1, PERSON2_SINGULAR, PERSON1_PLURAL,
  PERSON2_PLURAL
};

enum                            /* particle ze */
{
  ZE_PRESENT = 1
};

void lsym_clear(lsym *ls);
int lsym_isvalid(const lsym *ls);
int lsym_eq(const lsym *ls1, const lsym *ls2);
char *lsym_tostr(const lsym *ls, char s[]);
void lsym_genlist(FILE *f);
void lsym_ltest(void);
void lsym_print(const lsym *ls);

typedef struct dict dict;

typedef struct dictlist
{
  struct dictlist *next;
  dict *d;
} dictlist;

typedef struct dictdesc
{
  struct dictdesc *next;
  char mf[MAXLEN + 1];          /* main form */
  lsym ls;                      /* lexical symbol */
} dictdesc;

void dict_free(dict *d);
int dict_compile(FILE *f, FILE *g, int *line);
int dict_read(FILE *f, dict **d);
int dict_decompile(FILE *f, FILE *g);
int dict_find(const dict *d, const char *w, dictdesc **dd);

dictdesc *getsym(FILE *f, char *s, int *line, int *is_hanging,
                 const dictlist *dl);

int pla_run(int argc, char *argv[]);

#endif
